import meteva
import numpy as np
import matplotlib.pyplot as plt
import math
from matplotlib.colors import BoundaryNorm


def add_map(ax,add_county_line = False,add_worldmap = True,title = None,sup_fontsize = 12,linewidth = [0.3,0.3,0.2],color = ["k","k","k"]):

    if meteva.base.customized_basemap_list is None:
        if add_worldmap:
            meteva.base.tool.plot_tools.add_china_map_2basemap(ax, name="world", edgecolor='k', lw=0.3, encoding='gbk', grid0=None)  # "国界"
        meteva.base.tool.plot_tools.add_china_map_2basemap(ax, name="nation", edgecolor='k', lw=0.3,encoding = 'gbk')  #"国界"
        meteva.base.tool.plot_tools.add_china_map_2basemap(ax, edgecolor='k', lw=0.3,encoding = 'gbk')  #"省界"
        if add_county_line:
            meteva.base.tool.plot_tools.add_china_map_2basemap(ax, name="county", edgecolor='k', lw=0.2, encoding='gbk')  # "省界"
        ax.set_title(title,fontsize = sup_fontsize* 0.9)
        return ax
    else:
        for i in range(len(meteva.base.customized_basemap_list)):
            shpfile  = meteva.base.customized_basemap_list[i]
            # print(shpfile)
            encoding = 'utf-8'
            try:
                shp1 = meteva.base.tool.plot_tools.readshapefile(shpfile, default_encoding=encoding)
                lines = meteva.base.tool.plot_tools.LineCollection(shp1, antialiaseds=(1,), zorder=100)
                lines.set_color(color[i])
                lines.set_linewidth(linewidth[i])
                lines.set_label('_nolabel_')
                ax.add_collection(lines)
            except:
                encoding = "gbk"
                shp1 = meteva.base.tool.plot_tools.readshapefile(shpfile, default_encoding=encoding)
                lines = meteva.base.tool.plot_tools.LineCollection(shp1, antialiaseds=(1,), zorder=100)
                lines.set_color(color[i])
                lines.set_linewidth(linewidth[i])
                lines.set_label('_nolabel_')
                ax.add_collection(lines)


def creat_axs(nplot,map_extend,ncol = None,height  = None,width = None,dpi = 300,sup_title = None,sup_fontsize = 12,
              add_county_line = False,add_worldmap = True,add_minmap = None,title_list = None,add_index = None,wspace = None,grid = True,
              xticks_inter = None,yticks_inter = None,linewidth = [0.3,0.3,0.2],color = ["k","k","k"],keep_ticks = 1,
              width_colorbar = 0.5):


    ax_index = []
    if add_index is not None:
        if add_index is True:
            ax_index = ["(a)","(b)","(c)","(d)","(e)","(f)","(g)","(h)","(i)","(j)",
                        "(k)", "(l)", "(m)", "(n)", "(o)", "(p)", "(q)", "(r)", "(s)", "(t)",
                        "(u)", "(v)", "(w)", "(x)", "(y)", "(z)"]
        else:
            ax_index = add_index


    if isinstance(map_extend, list):
        slon = map_extend[0]
        elon = map_extend[1]
        slat = map_extend[2]
        elat = map_extend[3]
        rlon = elon - slon
        rlat = elat - slat
    else:
        slon = map_extend.slon
        slat = map_extend.slat
        elon = map_extend.elon
        elat = map_extend.elat
        rlon = map_extend.elon - map_extend.slon
        rlat = map_extend.elat - map_extend.slat


    if sup_title is None:
        sup_height_title = 0
    else:
        sup_height_title = sup_fontsize * 0.12

    height_title = sup_fontsize * 0.1
    if keep_ticks==1:
        height_hspace = sup_fontsize * 0.025
    else:
        height_hspace = sup_fontsize * 0.005

    if wspace is None:
        width_wspace = height_hspace*2
    else:
        width_wspace = wspace


    if keep_ticks==1:
        width_left_yticks = sup_fontsize * 0.2
        height_bottem_xticsk = sup_fontsize * 0.05
    else:
        width_left_yticks = 0
        height_bottem_xticsk =0

    if ncol is None:
        match_list = []
        for i in range(nplot, 0, -1):
            ncol = i
            nrow = int(math.ceil(nplot / ncol))
            rate = ncol * rlon / (nrow * rlat)
            if rate < 2 and rate > 9 / 16:
                match_list.append([i, ncol * nrow - nplot])
        if len(match_list) == 0:
            ncol = nplot
        else:
            match_array = np.array(match_list)
            min_index = np.argmin(match_array[:, 1])
            ncol = match_array[min_index, 0]
    nrow = int(math.ceil(nplot / ncol))

    if width is None and height is None:
        width = 8

    if width is None:
        height_all_plot = height - height_title - height_bottem_xticsk - (nrow - 1) * height_hspace + sup_height_title
        height_map = height_all_plot / nrow
        width_map = height_map * rlon / rlat
        width_all_plot = width_map * ncol + (ncol - 1) * width_wspace
        width = width_all_plot + width_colorbar + width_left_yticks
    else:
        width_all_plot = width - width_colorbar - width_left_yticks - (ncol - 1) * width_wspace
        width_map = width_all_plot / ncol
        height_map = width_map * rlat / rlon
        height_all_plot = height_map * nrow + (nrow - 1) * height_hspace
        height = height_all_plot + height_title + height_bottem_xticsk + sup_height_title

    vmax = elon
    vmin = slon
    if xticks_inter is None:
        r = rlon
        if r <= 0.1:
            inte = 0.05
        elif r <= 0.5:
            inte = 0.1
        elif r <= 1:
            inte = 0.2
        elif r <= 5 and r > 1:
            inte = 1
        elif r <= 10 and r > 5:
            inte = 2
        elif r < 20 and r >= 10:
            inte = 4
        elif r <= 30 and r >= 20:
            inte = 5
        elif r < 180:
            inte = 10
        else:
            inte = 20
    else:
        inte = xticks_inter

    vmin = inte * (math.ceil(vmin / inte))
    vmax = inte * ((int)(vmax / inte)+0.5)

    xticks = np.arange(vmin, vmax, inte)
    xticks_label = []
    xticks_label_None = []
    for x in range(len(xticks)):
        v1 = xticks[x]
        if abs(v1 - int(v1))<1e-7:
            xticks_label.append(str(int(round(v1, 6))))
        else:
            xticks_label.append(str(round(v1, 6)))
        xticks_label_None.append("")
    if xticks[-1] > 0:
        xticks_label[-1] = "   " + xticks_label[-1] + "°E"
    else:
        xticks_label[-1] = "   " + xticks_label[-1] + "°W"

    vmax = elat
    vmin = slat
    if yticks_inter is None:
        r = rlat
        if r <= 0.05:
            inte = 0.01
        elif r <= 0.3:
            inte = 0.05
        elif r <= 1:
            inte = 0.1
        elif r <= 5 and r > 1:
            inte = 1
        elif r <= 10 and r > 5:
            inte = 2
        elif r < 20 and r >= 10:
            inte = 4
        elif r <= 30 and r >= 20:
            inte = 5
        else:
            inte = 10
    else:
        inter = yticks_inter

    vmin = inte * (math.ceil(vmin / inte))
    vmax = inte * ((int)(vmax / inte)+0.5)
    yticks = np.arange(vmin, vmax, inte)
    yticks_label = []
    yticks_label_None = []
    for y in range(len(yticks)):
        v1 = yticks[y]
        if abs(v1 - int(v1))<1e-7:
            v1 = int(round(v1, 6))
        else:
            v1 = round(v1, 6)

        if yticks[y] >= 0:
            yticks_label.append(str(v1) + "°N")
        else:
            yticks_label.append(str(v1) + "°S")
        yticks_label_None.append("")

    fig = plt.figure(figsize=(width, height), dpi=dpi)
    y_sup_title = (height_bottem_xticsk  + nrow * height_map +(nrow-1) * height_hspace) / height
    if sup_title is not None:
        plt.suptitle(sup_title, x = width_left_yticks / width,y = y_sup_title,fontsize=sup_fontsize,horizontalalignment ="left",
                     verticalalignment  ="bottom")
    ax_list = []
    min_ax_list = []
    for p in range(nplot):
        pi = p % ncol
        pj = int(p / ncol)

        rect1 = [(width_left_yticks + pi * (width_map + width_wspace)) / width,
                 (height_bottem_xticsk + (nrow - 1 - pj) * (height_map + height_hspace)) / height,
                 width_map / width,
                 height_map / height]
        ax = plt.axes(rect1)
        ax.set_xlim((slon, elon))
        ax.set_ylim((slat, elat))

        if keep_ticks>=0:
            if p >= nplot - ncol or keep_ticks:
                # 最底行画横坐标
                ax.set_xticks(xticks)
                ax.set_xticklabels(xticks_label, fontsize=sup_fontsize * 0.9) #, family='Times New Roman')
            else:
                ax.set_xticklabels("", fontsize=sup_fontsize * 0.9) #, family='Times New Roman')


            if pi ==0 or keep_ticks:
                # 最左侧画纵坐标
                ax.set_yticks(yticks)
                ax.set_yticklabels(yticks_label, fontsize=sup_fontsize * 0.9) #, family='Times New Roman')
            else:
                ax.set_yticklabels("", fontsize=sup_fontsize * 0.9) #, family='Times New Roman')
        else:
            ax.set_xticks([])
            ax.set_yticks([])
        if grid: plt.grid(linestyle = "--",linewidth = 0.5)
        if title_list is None:
            sub_title = None
        else:
            sub_title = title_list[p]
        add_map(ax,add_county_line=add_county_line,add_worldmap=add_worldmap,sup_fontsize=sup_fontsize,title=sub_title,
                linewidth = linewidth,color =color)

        if len(ax_index)>=1:
            ix = slon + 0.02*(elon - slon) * 5 / width_map
            iy = elat - 0.035*(elat - slat) * 5 / height_map
            plt.text(ix, iy, ax_index[p], bbox=dict(fc='white', ec='white',pad = 0),fontsize = sup_fontsize,zorder=100)
        ax_list.append(ax)


        if slon< 75 and elon >130 and elat >50 and slat >3 and slat <25:
            if add_minmap is None:
                print("返回结果中自动添加了南海小地图，请注意返回结果中包含了两个绘图框列表")
                add_minmap = "left"

        if add_minmap is not None:
            if add_minmap == "left" or add_minmap=="right":
                minmap_lon_lat = [105, 123, 0, 20]
                minmap_height_rate = 0.27
                height_bigmap = rect1[3]
                height_minmap = height_bigmap * minmap_height_rate
                width_minmap = height_minmap * (minmap_lon_lat[1] - minmap_lon_lat[0]) * height / (
                            minmap_lon_lat[3] - minmap_lon_lat[2])/width

                width_between_two_map =  height_bigmap *0.01
                sy_minmap = width_between_two_map + rect1[1]
                if add_minmap =="left":
                    sx_minmap = rect1[0] + width_between_two_map
                else:
                    sx_minmap = rect1[0] + rect1[2] - width_minmap - width_between_two_map
                rect_min = [sx_minmap,sy_minmap,width_minmap,height_minmap]
                ax_min = plt.axes(rect_min)
                plt.xticks([])
                plt.yticks([])
                ax_min.set_xlim((minmap_lon_lat[0], minmap_lon_lat[1]))
                ax_min.set_ylim((minmap_lon_lat[2], minmap_lon_lat[3]))
                ax_min.spines["top"].set_linewidth(0.3)
                ax_min.spines["bottom"].set_linewidth(0.3)
                ax_min.spines["right"].set_linewidth(0.3)
                ax_min.spines["left"].set_linewidth(0.3)
                add_map(ax_min, add_worldmap =True)  # "国界"
                min_ax_list.append(ax_min)
    if len(min_ax_list)==0:
        return ax_list
    else:
        return ax_list,min_ax_list



def add_contourf(ax,grd,cmap ="rainbow",clevs= None,add_colorbar = True,cut_colorbar = True,title = None,title_fontsize = 8,clip = None,
                 extend = None,colorbar_location = None,alpha=1):
    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1
    if grd is None:return
    grid0 = meteva.base.get_grid_of_data(grd)
    grid1 = meteva.base.grid([slon,elon,grid0.dlon],[slat,elat,grid0.dlat])
    grd1 = meteva.base.interp_gg_linear(grd,grid1,outer_value=np.nan)

    x = grd1['lon'].values
    y = grd1['lat'].values
    vmax = np.nanmax(grd1.values)
    vmin = np.nanmin(grd1.values)
    if extend is None: extend = "neither"
    cmap1,clevs1 = meteva.base.tool.color_tools.def_cmap_clevs(cmap=cmap,clevs=clevs,vmin=vmin,vmax = vmax,cut_colorbar = cut_colorbar,extend =extend)
    # if extend == "neither":
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N - 1)
    # elif extend == "both":
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N + 1)
    # else:
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N)
    norm = BoundaryNorm(clevs1, ncolors=cmap1.N)
    im = ax.contourf(x, y, np.squeeze(grd1.values), levels=clevs1, cmap=cmap1,norm = norm,alpha=alpha,extend = extend)
    fig = plt.gcf()
    width = fig.bbox.width/fig.dpi
    height = fig.bbox.height/fig.dpi

    ax.set_title(title,fontsize =title_fontsize)
    if add_colorbar:
        if colorbar_location is None:
            location = [ax.bbox.x1 / fig.dpi / width + 0.005, ax.bbox.y0 / fig.dpi / height, 0.01,
                        ax.bbox.height / fig.dpi / height]
        else:
            location = colorbar_location
        colorbar_position = fig.add_axes(location)  # 位置[左,下,宽,高]
        plt.colorbar(im,cax= colorbar_position)

    if clip is not None:
        try:
            meteva.base.tool.maskout.shp2clip_by_shpfile(im,ax,clip)
        except:
            if isinstance(clip, str): clip = [clip]
            if isinstance(clip[0], str):
                meteva.base.tool.maskout.shp2clip_by_region_name(im, ax, clip)
            else:
                meteva.base.tool.maskout.shp2clip_by_lines(im, ax, clip)

    return im

def add_contour(ax,grd,color='k', linewidth = 1,label_fontsize = 5,clevs = None,title = None,title_fontsize = 8,clip = None):


    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1

    grid0 = meteva.base.get_grid_of_data(grd)
    grid1 = meteva.base.grid([slon,elon,grid0.dlon],[slat,elat,grid0.dlat])
    grd1 = meteva.base.interp_gg_linear(grd,grid1,outer_value=np.nan)

    x = grd1['lon'].values
    y = grd1['lat'].values
    vmax = np.nanmax(grd1.values)
    vmin = np.nanmin(grd1.values)


    if clevs is None:
        dif = (vmax - vmin) / 10.0
        inte = math.pow(10, math.floor(math.log10(dif)));
        #用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
        r = dif / inte
        if r < 3 and r >= 1.5:
            inte = inte * 2
        elif r < 4.5 and r >= 3:
            inte = inte * 4
        elif r < 5.5 and r >= 4.5:
            inte = inte * 5
        elif r < 7 and r >= 5.5:
            inte = inte * 6
        elif r >= 7:
            inte = inte * 8
        vmin = inte * (math.floor(vmin / inte))
        vmax = inte * (math.ceil(vmax / inte) + 0.5)
        clevs = np.arange(vmin, vmax, inte)
    mm = max(abs(clevs[0]),abs(clevs[-1]))
    if mm!=0:
        mmi = math.ceil(math.log10(mm))
        if mmi >=2:
            f1 = mmi
            f2 = 0
        elif mmi == 1:
            f1 = 2
            f2 = 1
        else:
            mmi = math.floor(math.log10(mm))
            f1 = 1+abs(mmi)
            f2 = f1
    else:
        f1 = 1
        f2 = 0
    fmt = "%"+str(f1)+"."+str(f2)+"f"
    ax.set_title(title,fontsize =title_fontsize)
    im = ax.contour(x, y, np.squeeze(grd1.values),levels = clevs,colors = color,linewidths = linewidth)
    ax.clabel(im, inline=1, fontsize=label_fontsize,fmt=fmt)
    if clip is not None:
        try:
            meteva.base.tool.maskout.shp2clip_by_shpfile(im,ax,clip)
        except:
            if isinstance(clip, str): clip = [clip]
            if isinstance(clip[0], str):
                meteva.base.tool.maskout.shp2clip_by_region_name(im, ax, clip)
            else:
                meteva.base.tool.maskout.shp2clip_by_lines(im, ax, clip)

    return im



def add_mesh(ax,grd,cmap ="rainbow",clevs= None,add_colorbar = True,title = None,title_fontsize = 8,extend = None,colorbar_location = None):


    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1

    grid0 = meteva.base.get_grid_of_data(grd)
    grid1 = meteva.base.grid([slon,elon,grid0.dlon],[slat,elat,grid0.dlat])
    grd1 = meteva.base.interp_gg_linear(grd,grid1,outer_value=np.nan)

    x = grd1['lon'].values
    y = grd1['lat'].values
    vmax = np.nanmax(grd1.values)
    vmin = np.nanmin(grd1.values)
    if extend is None: extend="neither"
    cmap1,clevs1 = meteva.base.tool.color_tools.def_cmap_clevs(cmap=cmap,clevs=clevs,vmin=vmin,vmax = vmax,extend=extend)
    # if extend is None:
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N - 1)
    # elif extend == "both":
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N)
    #     #norm = BoundaryNorm(clevs1, ncolors=cmap1.N + 1, extend=extend)
    # else:
    #     #norm = BoundaryNorm(clevs1, ncolors=cmap1.N,extend = extend)
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N)
    norm = BoundaryNorm(clevs1, ncolors=cmap1.N)
    im = ax.pcolormesh(x, y, np.squeeze(grd1.values), cmap=cmap1, norm=norm)
    #im = ax.contourf(x, y, np.squeeze(grd.values), levels=clevs1, cmap=cmap1,norm = norm)

    fig = plt.gcf()

    #left_low = (width +0.1 - 0.8) / width

    ax.set_title(title,fontsize =title_fontsize)
    if add_colorbar:
        if colorbar_location is None:
            width = fig.bbox.width / fig.dpi
            height = fig.bbox.height / fig.dpi
            location = [ax.bbox.x1 / fig.dpi / width + 0.005, ax.bbox.y0 / fig.dpi / height, 0.01,
                        ax.bbox.height / fig.dpi / height]
        else:
            location = colorbar_location

        colorbar_position = fig.add_axes(location)  # 位置[左,下,宽,高]
        plt.colorbar(im,cax= colorbar_position,extend = extend)
    return im

def add_barbs(ax,wind,color = "k",skip = None,title = None,title_fontsize = 8,length = None):

    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1

    grid0 = meteva.base.get_grid_of_data(wind)
    grid1 = meteva.base.grid([slon,elon,grid0.dlon],[slat,elat,grid0.dlat])
    wind1 = meteva.base.interp_gg_linear(wind,grid1,outer_value=np.nan)

    x = wind1['lon'].values
    y = wind1['lat'].values
    X, Y = np.meshgrid(x, y)
    u = np.squeeze(wind1.values[0,...])
    v = np.squeeze(wind1.values[1,...])

    fig = plt.gcf()
    width = fig.bbox.width/fig.dpi

    if skip is  not None:
        if length is None:
            length = math.sqrt((width-1) * skip / x.size) * 8
    else:
        if length is None:
            length = ax.bbox.width/fig.dpi
        skip =int((length /10)**2 * x.size / (width - 1))+1
    ax.set_title(title,fontsize =title_fontsize)
    im = ax.barbs(X[::skip,::skip],Y[::skip,::skip] , u[::skip,::skip], v[::skip,::skip],
             sizes=dict(emptybarb=0.01, spacing=0.23, height=0.5,width = 0.25),color = color,
             barb_increments=dict(half=2, full=4, flag=20),length = length,linewidth = length * length  * 0.03)


    return im


def add_scatter(ax,sta0,cmap = "rainbow",clevs = None,point_size = None,fix_size = True,threshold = 2,min_spot_value = 0,mean_value = 2,
                add_colorbar = True,alpha = None,title = None,title_fontsize = 8,extend = None,colorbar_location = None):

    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1
    rlon = elon - slon
    rlat = elat - slat
    sta = sta0

    sta_without_iv = meteva.base.sele.not_IV(sta)
    data_names = meteva.base.get_stadata_names(sta_without_iv)
    sta_without_iv = sta_without_iv.sort_values(by=data_names[-1], ascending=True)

    values = sta_without_iv.iloc[:, -1].values

    vmax_v = np.max(sta_without_iv.iloc[:,-1].values)
    vmin_v = np.min(sta_without_iv.iloc[:,-1].values)
    if extend is None: extend="neither"
    cmap1, clevs1 = meteva.base.tool.color_tools.def_cmap_clevs(cmap=cmap, clevs=clevs, vmin=vmin_v, vmax=vmax_v,extend = extend)
    # if extend is None:
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N - 1)
    # elif extend == "both":
    #     #norm = BoundaryNorm(clevs1, ncolors=cmap1.N + 2, extend=extend)
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N)
    # else:
    #     #norm = BoundaryNorm(clevs1, ncolors=cmap1.N+1, extend=extend)
    #     norm = BoundaryNorm(clevs1, ncolors=cmap1.N)
    norm = BoundaryNorm(clevs1, ncolors=cmap1.N)
    fig = plt.gcf()
    map_width = ax.bbox.width/fig.dpi

    if point_size is None:
        sta_id1 = sta0.drop_duplicates(['id'])
        sta_dis = meteva.base.sta_dis_ensemble_near_by_sta(sta_id1, nearNum=2)
        dis_values = sta_dis["data1"].values
        dis_values.sort()
        dis1 = dis_values[int(len(dis_values) * 0.02) + 1]/1.2
        point_size = (map_width * dis1 / rlon) ** 2
        # point_size = 100 * map_area / len(sta.index)
        # print("**************")
        # print(point_size)
        if (point_size > 30): point_size = 30
        if (point_size < 0.5):
            point_size = 0.5
            if alpha is None:
                alpha = 0.5
        # point_size *=3
        #left_low = (width + 0.1 - right_plots_width) / width


    x = sta_without_iv.loc[:, "lon"].values
    y = sta_without_iv.loc[:, "lat"].values
    colors = values
    if isinstance(fix_size, bool):
        if fix_size:
            im = ax.scatter(x, y, c=colors, cmap=cmap1, norm=norm, s=point_size,edgecolors = "face",alpha = alpha)
        else:
            area = point_size * np.abs(values - min_spot_value) / mean_value
            if (threshold is not None):
                area[np.abs(values - min_spot_value) < threshold] *= 0.1
            im = ax.scatter(x, y, c=colors, cmap=cmap1, norm=norm, s=area,edgecolors  = 'face',alpha = alpha)

    else:
        im = ax.scatter(x, y, c=colors, cmap=cmap1, norm=norm, s=fix_size,edgecolors  = 'face',alpha = alpha)
    ax.set_title(title,fontsize =title_fontsize)


    if add_colorbar:
        if colorbar_location is None:
            width = fig.bbox.width / fig.dpi
            height = fig.bbox.height / fig.dpi
            location = [ax.bbox.x1 / fig.dpi / width + 0.005, ax.bbox.y0 / fig.dpi / height, 0.01,
                        ax.bbox.height / fig.dpi / height]
        else:
            location = colorbar_location

        colorbar_position = fig.add_axes(location)  # 位置[左,下,宽,高]
        plt.colorbar(im, cax=colorbar_position,extend = extend)
    return im

def add_scatter_text(ax,sta0,color = "k",cmap = None,clevs = None,tag = 2,
               alpha = 1,font_size = 10,title = None,title_fontsize = 8):
    x0 = ax.transLimits._boxin.x0
    x1 = ax.transLimits._boxin.x1
    y0 = ax.transLimits._boxin.y0
    y1 = ax.transLimits._boxin.y1
    sta = meteva.base.sele_by_para(sta0,lon=[x0,x1],lat=[y0,y1])
    sta_without_iv = meteva.base.sele.not_IV(sta)
    data_names = meteva.base.get_stadata_names(sta_without_iv)
    sta_without_iv = sta_without_iv.sort_values(by=data_names[-1], ascending=True)

    vmax_v = np.max(sta_without_iv.iloc[:,-1].values)
    vmin_v = np.min(sta_without_iv.iloc[:,-1].values)
    clevs1 = None
    cmap1 = None
    if cmap is not None:
        cmap1, clevs1 = meteva.base.tool.color_tools.def_cmap_clevs(cmap=cmap, clevs=clevs, vmin=vmin_v, vmax=vmax_v)
    fig = plt.gcf()
    fmt_tag = "%." + str(tag) + "f"

    if cmap1 is None:
        nsta = len(sta_without_iv.index)
        for i in range(nsta):
            x = sta_without_iv.iloc[i,4]
            y = sta_without_iv.iloc[i,5]
            v = sta_without_iv.iloc[i,-1]
            if isinstance(v,str):
                ax.text(x, y, v, ha="center", va="center", fontsize=font_size, color=color, clip_on=True,alpha = alpha, zorder=1000)
            else:
                ax.text(x, y, fmt_tag % v, ha="center", va="center",fontsize=font_size,color = color, clip_on=True,alpha = alpha, zorder=1000)
    else:
        x = sta_without_iv.iloc[:, 4].values
        y = sta_without_iv.iloc[:, 5].values
        v = sta_without_iv.iloc[:, -1].values
        for k in range(len(clevs1)-1):
            index = np.where((v>=clevs1[k])&(v<clevs1[k+1]))
            if len(index[0])>0:
                x1 = x[index]
                y1 = y[index]
                v1 = v[index]
                color = cmap1(k)
                for j in range(x1.size):
                    if isinstance(v1[j],str):
                        ax.text(x1[j], y1[j], v1[j], ha="center", va="center", fontsize=font_size, c=color, clip_on=True,alpha = alpha, zorder=1000)
                    else:
                        ax.text(x1[j], y1[j], fmt_tag % v1[j], ha="center", va="center", fontsize=font_size,  c=color, clip_on=True,alpha = alpha, zorder=1000)

    ax.set_title(title,fontsize =title_fontsize)
    return

def add_closed_line(ax,graphy,color = "k",linewidth=2,fontsize = 10,title = None,title_fontsize = 8):
    if graphy is None:
        return
    x0 = ax.transLimits._boxin.x0
    x1 = ax.transLimits._boxin.x1
    y0 = ax.transLimits._boxin.y0
    y1 = ax.transLimits._boxin.y1

    contours =graphy["closed_contours"]
    ncontour = len(contours["cn_label"])


    for n in range(ncontour):
        line_array = contours["cn_xyz"][n][:,0:2]
        ax.plot(line_array[:,0], line_array[:,1], color=color, linewidth=linewidth)
        value = contours["cn_label"][n]
        i_s = -1
        line_dict = None
        while i_s < line_array.shape[0]-1:
            i_s += 1
            x = line_array[i_s,0]
            y = line_array[i_s,1]
            if x >=x0 and x<= x1 and y >=y0 and y <= y1:
                if i_s == 0:
                    x_1 = line_array[1,0]
                    y_1 = line_array[1,1]
                    if x_1 == x:
                        rotation = 90
                    else:
                        rotation = math.atan((y_1 - y)/(x_1 - x)) * 180 / math.pi
                    ax.text(x, y, value, ha="center", va="center", fontsize=fontsize, color=color,
                            rotation = rotation,bbox ={"facecolor":"w","pad":0,"linewidth":0})
                    break
                else:
                    if line_dict is None:
                        line_dict = {}
                        line_dict["x"] = []
                        line_dict["y"] = []
                    else:
                        line_dict["x"].append(x)
                        line_dict["y"].append(y)
            else:
                if line_dict is not None:
                    np = len(line_dict["x"])
                    if np >5:
                        np_half = int(np/2)
                        x = line_dict["x"][np_half]
                        y = line_dict["y"][np_half]
                        x_1 =  line_dict["x"][np_half+1]
                        y_1 = line_dict["y"][np_half+1]
                        if x_1 == x:
                            rotation = 90
                        else:
                            rotation = math.atan((y_1 - y) / (x_1 - x)) * 180 / math.pi
                        ax.text(x, y, value, ha="center", va="center", fontsize=fontsize, color=color,
                                 rotation=rotation,bbox ={"facecolor":"w","pad":0,"linewidth":0})
                        line_dict = None
    ax.set_title(title,fontsize =title_fontsize)
    return


def add_shear_line(ax,graphy,linewidth = 1,title = None,title_fontsize = 8):
    if graphy is None:
        return
    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1

    rlon = elon - slon
    rlat = elat - slat
    fig = plt.gcf()
    map_width = ax.bbox.width / fig.dpi

    line_width = rlon *  0.015/ map_width

    features = graphy["features"]
    for value in features.values():
        line = value["axes"]
        line_type = line["line_type"]
        point = np.array(line["point"])
        dp = np.zeros(point.shape)
        dp[:-1,:] = point[1:,:] - point[:-1,:]
        dp[-1,:] = dp[-2,:]
        length = (dp[:,0]**2 + dp[:,1] **2)**0.5
        dx = -dp[:,1] * line_width/ length
        dy = dp[:,0] * line_width/length


        if line_type =="c":
            ax.plot(point[:,0] - dx ,point[:,1] - dy,"b",linewidth = linewidth)
            ax.plot(point[:, 0] + dx, point[:, 1] + dy, "b",linewidth =  linewidth)
        else:
            ax.plot(point[:, 0] - dx, point[:, 1] - dy, "r",linewidth =  linewidth)
            ax.plot(point[:, 0] + dx, point[:, 1] + dy, "r",linewidth =  linewidth)
    ax.set_xlim(slon,elon)
    ax.set_ylim(slat,elat)
    ax.set_title(title,fontsize =title_fontsize)
    return

def add_trough_axes(ax,graphy,color = "r",linewidths = None,title = None,title_fontsize = 8):
    if graphy is None:
        return
    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1

    rlon = elon - slon
    rlat = elat - slat
    fig = plt.gcf()
    map_width = ax.bbox.width / fig.dpi

    if linewidths is None:
        linewidths = rlon *  0.015/ map_width

    features = graphy["features"]
    for value in features.values():
        line = value["axes"]
        point = np.array(line["point"])
        if point.size>2:
            ax.plot(point[:, 0], point[:, 1],color, linewidth=linewidths)

    ax.set_xlim(slon,elon)
    ax.set_ylim(slat,elat)
    ax.set_title(title,fontsize =title_fontsize)
    return

def add_jet_axes(ax,graphy,title = None,title_fontsize = 8,color ="yellow" ):
    if graphy is None:
        return
    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1

    rlon = elon - slon
    rlat = elat - slat
    fig = plt.gcf()
    map_width = ax.bbox.width / fig.dpi

    line_width = rlon *  0.015/ map_width

    features = graphy["features"]
    for value in features.values():
        line = value["axes"]
        point = np.array(line["point"])
        npoint = len(line["point"])
        if npoint>2:
            ns = npoint -1
            dx = 0
            dy = 0
            while ns >0:
                ns -= 1
                dx = point[npoint-1,0] - point[ns,0]
                dy = point[npoint-1,1] - point[ns,1]
                dis = (dx**2 + dy**2)**0.5
                if dis > 0.3:
                    break

            ax.arrow(point[ns-1,0],point[ns-1,1],dx*0.01,dy*0.01,head_width=1,head_length = 1,fc = color,ec = color)
            ax.plot(point[:ns, 0], point[:ns, 1], color, linewidth=2.5)

    ax.set_xlim(slon,elon)
    ax.set_ylim(slat,elat)
    ax.set_title(title,fontsize =title_fontsize)
    return

def add_cyclone_trace(ax,sta_cyclone_trace,size = 0.3,linewidth = 1,title = None,title_fontsize = 8):

    np = len(sta_cyclone_trace.index)
    for i in range(np - 1):
        lon = [sta_cyclone_trace.iloc[i,4],sta_cyclone_trace.iloc[i+1,4]]
        lat = [sta_cyclone_trace.iloc[i,5],sta_cyclone_trace.iloc[i+1,5]]
        speed = sta_cyclone_trace.iloc[i+1,-1]


        if speed >= 10.8 and speed <17.2:
            color = "k"
        elif speed >=17.2 and speed <24.5:
            color = "b"
        elif speed >= 24.5 and speed < 32.7:
            color = "y"
        elif speed >= 32.7 and speed < 41.5:
            color = "orange"
        elif speed >=41.5 and speed < 51:
            color = "r"
        else:
            color = "m"
        ax.plot(lon, lat, c = color, linewidth=linewidth, zorder=29)
        time_ob = sta_cyclone_trace.iloc[i,1]

        if (time_ob.hour - 2) %12 ==0:
            r_o = size
            lon_list = []
            lat_list = []
            for a in range(0,360,5):
                if a> 70 and a<135:
                    lon1 = lon[0] + r_o *2.5 * math.cos(70 * math.pi/180)
                    lat1 = lat[0] + r_o *2.5 * math.sin(70 * math.pi/180)
                elif a >250 and a < 315:
                    lon1 = lon[0] + r_o * 2.5 * math.cos(250 * math.pi / 180)
                    lat1 = lat[0] + r_o * 2.5 * math.sin(250 * math.pi / 180)
                else:
                    theta = math.pi * a / 180
                    lon1 = lon[0] + r_o * math.cos(theta)
                    lat1 = lat[0] + r_o * math.sin(theta)
                lon_list.append(lon1)
                lat_list.append(lat1)
            ax.fill(lon_list,lat_list,c = color,zorder = 30,linewidth=0)
            r_o = size * 0.3
            lon_list = []
            lat_list = []
            for a in range(0,360,5):
                theta = math.pi * a / 180
                lon1 = lon[0] + r_o * math.cos(theta)
                lat1 = lat[0] + r_o * math.sin(theta)
                lon_list.append(lon1)
                lat_list.append(lat1)
            ax.fill(lon_list,lat_list,c = "white",zorder = 30,linewidth=0)
    ax.set_title(title,fontsize =title_fontsize)
    return


def add_lines(ax,graphy,cent = None,color = "k",linestyle = "-",linewidths = None,title = None,title_fontsize = 8):
    if graphy is None:
        return
    slon = ax.transLimits._boxin.x0
    elon = ax.transLimits._boxin.x1
    slat = ax.transLimits._boxin.y0
    elat = ax.transLimits._boxin.y1

    rlon = elon - slon
    rlat = elat - slat
    fig = plt.gcf()
    map_width = ax.bbox.width / fig.dpi

    if linewidths is None:
        linewidths = rlon *  0.015/ map_width

    if isinstance(graphy,dict):
        features = graphy["features"]
        for value in features.values():
            line = value["axes"]
            point = np.array(line["point"])
            ax.plot(point[:, 0], point[:, 1], color, linewidth=linewidths,linestyle=linestyle)
    else:
        df_list = meteva.base.split(graphy,used_coords=["level","time","dtime","id"])
        for df in df_list:
            lon = df["lon"].values
            lat = df["lat"].values
            ax.plot(lon,lat, color, linewidth=linewidths,linestyle = linestyle)

            if cent is not None:
                id = df["id"].values[0]
                level = df["level"].values[0]
                time = df["time"].values[0]
                dtime = df["dtime"].values[0]
                cent1 = meteva.base.sele_by_para(cent,level = level,time = time,dtime = dtime,id = id)



    ax.set_xlim(slon,elon)
    ax.set_ylim(slat,elat)
    ax.set_title(title,fontsize =title_fontsize)
    return




import numpy as np
import scipy.stats as ss
import math
from .fft2d import fft2d
from .cbind import cbind
import numpy.fft
#import pyreadr


def aaft2d(im, bigdim=None):
    im = np.array(im)
    xdim = im.shape
    im_vec = im.flatten()
    n = im_vec.size
    im2 = np.random.normal(size=xdim)
    im2_vec = im2.flatten()
    out = np.full(xdim, np.nan)
    rk1 = ss.rankdata(im_vec) - 1
    s_im2_vec = np.sort(im2_vec)
    rk1 = np.floor(rk1).astype(int)
    z = s_im2_vec[rk1]
    if bigdim is None:
        if xdim[0] <= 1024:
            x_index = 2 ** math.ceil(math.log2(xdim[0]))
        else:
            x_index = math.ceil(xdim[0] / 512) * 512
        if xdim[1] <= 1024:
            y_index = 2 ** math.ceil(math.log2(xdim[1]))
        else:
            y_index = math.ceil(xdim[1] / 512) * 512
        bigdim = (x_index, y_index)
    big_0_mat = np.full(bigdim, 0 + 0j)
    big_n = np.prod(bigdim)
    x = np.full(bigdim, 0.0)
    x[0: xdim[0], 0: xdim[1]] = np.array(z).reshape(xdim)
    zfft = fft2d(x, bigdim=bigdim)
    nr1 = math.ceil(xdim[0] / 2) - 1
    nr2 = (xdim[0] - 1)
    nc2 = math.ceil(xdim[1] / 2) - 1
    pi = 2 * math.pi * np.random.normal(size=(nr1, 1))
    pj = 2 * math.pi * np.random.normal(size=(1, nc2))
    pij = 2 * math.pi * np.random.normal(size=(nr2, nc2))
    if xdim[1] % 2 == 0:
        qij = 2 * math.pi * np.random.normal(size=(nr1, 1))
        if xdim[0] % 2 == 0:
            qij = np.concatenate((qij, np.zeros(1), -np.flipud(qij)))
        else:
            qij = np.concatenate((qij, -np.flipud(qij)))
        qij_shape = (nr2, qij.size / nr2)
        qij = qij.reshape(qij_shape)
        phij = cbind(cbind(pij, qij), -np.fliplr(np.flipud(pij)))
    else:
        phij = cbind(pij, -np.fliplr(np.flipud(pij)))
    if xdim[0] % 2 == 0:
        phi = np.concatenate((np.zeros(1), pi.flatten(), np.zeros(1), -np.flipud(pi).flatten()))
    else:
        phi = np.concatenate((np.zeros(1), pi.flatten(), -np.flipud(pi).flatten()))
    if xdim[1] % 2 == 0:
        phj = np.concatenate((np.zeros(1), pj.flatten(), np.zeros(1), -np.fliplr(pj).flatten()))
    else:
        phj = np.concatenate((np.zeros(1), pj.flatten(), -np.fliplr(pj).flatten()))
    ph = np.full(xdim, 0.0)
    ph[1: xdim[0], 1: xdim[1]] = phij
    ph[0, :] = phj
    ph[:, 0] = phi
    k = big_0_mat
    k[0: xdim[0], 0: xdim[1]] = zfft[0: xdim[0], 0: xdim[1]] * np.exp((0 + 1j) * ph)
    j = np.array(np.real(np.fft.ifft(k.flatten()) * k.size / n).reshape(k.shape)[0: xdim[0], 0: xdim[1]])
    rk2 = ss.rankdata(j.flatten()) - 1
    j = np.sort(im_vec)
    rk2 = np.floor(rk2).astype(int)
    j = np.array(j[rk2])
    out = j.reshape(xdim)
    return out


if __name__ == '__main__':
    # geom000 = pyreadr.read_r('../data/geom000.Rdata')['geom000']
    # z = aaft2d(geom000)
    # print("h")
    pass



import meteva
import numpy as np
import time
import sys
import copy
from meteva.method.space.mode import feature_finder
from meteva.method.space.mode.centmatch import centmatch
from meteva.method.space.mode.merge_force import merge_force
from meteva.method.space.mode import data_pre
from meteva.method.space.rigider import rigider


def get_rigid(x, ob_i,fo_j,stages,translate,rotate):
    '''
    调用刚体变换，将一个预报目标向通过平移和旋转至其匹配的观测目标的位置
    :param x:
    :param ob_i:
    :param fo_j:
    :param stages:
    :param translate:
    :param rotate:
    :return:
    '''
    grid0 = x["grid"] #网格

    grd_ob_i = meteva.base.grid_data(grid0)
    index = x["grd_ob_label"].values==ob_i
    grd_ob_i.values[index] =  x["grd_ob"].values[index]   #第i个观测目标区域保持要素值，其它区域清零

    grd_fo_j = meteva.base.grid_data(grid0)
    index = x["grd_fo_label"].values==fo_j
    grd_fo_j.values[index] =  x["grd_fo"].values[index]  #第j个预报目标区域保持要素值，其它区域清零

    look_rigid = rigider.rigid_optimal(grd_ob_i,grd_fo_j,stages=stages,translate=translate,rotate=rotate)  #通过刚体变换，找到最佳平移、旋转角度，以及变换后的预报场

    return look_rigid


def caculate_error_components(x, translate, rotate):
    '''
    根据刚体变换的结果计算预报误差的不同分量
    :param x:
    :param translate:
    :param rotate:
    :return:
    '''
    grd_fo = x['grd_fo']
    grd_ob = x['grd_ob']
    if translate:
        if rotate:
            grd_fo_shift = x['grd_fo_shift']  # rigider没有输出该结果
            grd_fo_shift_rotate = x['grd_fo_shift_rotate']
        else:
            grd_fo_shift = x['grd_fo_shift']
            grd_fo_shift_rotate = grd_fo_shift
    else:
        if rotate:
            grd_fo_shift = grd_fo
            grd_fo_shift_rotate = x['grd_fo_shift_rotate']
        else:
            print("刚体变换的translate 和rotate不能同时为False")
            return None

    values = np.abs(grd_ob.values) + np.abs(grd_fo.values) +np.abs(grd_fo_shift.values)+ np.abs(grd_fo_shift_rotate.values)
    index = np.where(values!=0)
    imax = np.max(index[-1])+1
    imin = np.min(index[-1])
    jmax = np.max(index[-2])+1
    jmin = np.min(index[-2])

    grd_ob_values = grd_ob.values[0,0,0,0,jmin:jmax,imin:imax]
    grd_fo_values = grd_fo.values[0,0,0,0,jmin:jmax,imin:imax]
    grd_fo_shift_values = grd_fo_shift.values[0,0,0,0,jmin:jmax,imin:imax]
    grd_fo_shift_rotate_values = grd_fo_shift_rotate.values[0,0,0,0,jmin:jmax,imin:imax]

    MSE_total = np.mean(np.power(grd_fo_values - grd_ob_values, 2))  # 总误差
    d2_shift_left = np.power(grd_fo_shift_values -grd_ob_values, 2)  #平移场剩余误差场
    MSE_shift_left = np.mean(d2_shift_left)                            # 平移剩余误差
    MSE_shift = MSE_total - MSE_shift_left                     #平移误差

    d2_shift_rotate_left = np.power(grd_fo_shift_rotate_values - grd_ob_values, 2)  #刚体变换后误差场
    MSE_shift_rotate_left = np.mean(d2_shift_rotate_left)                             #刚体变换后的均方误差


    fo_mean  = np.mean(grd_fo_values)     # 平移旋转后观测预报并集区域的均值。
    ob_mean  = np.mean(grd_ob_values)     # 平移旋转后观测预报并集区域的均值。
    MSE_volume = np.power(fo_mean - ob_mean, 2)


    res = {"ob_mean":ob_mean,"fo_mean":fo_mean,
        "MSE_total": MSE_total, "MSE_shift_left": MSE_shift_left,
           "MSE_shift": MSE_shift, "MSE_shift_rotate_left": MSE_shift_rotate_left,
           "MSE_rotate": MSE_shift_left - MSE_shift_rotate_left, "MSE_volume": MSE_volume,
           "MSE_pattern": MSE_shift_rotate_left - MSE_volume}

    return res

def craer(look_merge,stages=True,translate=True,rotate=True):

    out = {}
    nmatch = len(look_merge["matches"])
    for k in range(nmatch):
        ob_i = look_merge["matches"][k][0]
        fo_j = look_merge["matches"][k][1]
        look_rigid = get_rigid(look_merge,ob_i,fo_j,stages,translate,rotate)
        error_components = caculate_error_components(look_rigid,stages,rotate)
        for key in error_components.keys():
            look_rigid[key] = error_components[key]
        out[ob_i] = look_rigid

    return out

def operation(grd_ob,grd_fo,smooth,threshold,minsize,compare = ">=", stages = True,translate = True,rotate = True):
    look_featureFinder = feature_finder(grd_ob,grd_fo,smooth=smooth,threshold=threshold,minsize=minsize,compare=compare)
    look_match = centmatch(look_featureFinder)
    look_merge = merge_force(look_match)
    look_cra = craer(look_merge,stages,translate,rotate)
    result = {"merge":look_merge,
              "cra":look_cra}
    return result

def plot_value(look_cra,label,cmap = "rain_1h",clevs= None):

    rire = look_cra[label]
    grd_list = [rire["grd_ob"],rire["grd_fo"],rire["grd_fo_shift"],rire["grd_fo_shift_rotate"]]
    vmax = max(np.max(grd_list[0].values),np.max(grd_list[0].values))
    vmin = min(np.min(grd_list[0].values),np.min(grd_list[0].values))
    cmap1, clevs1 = meteva.base.tool.color_tools.def_cmap_clevs(cmap=cmap, clevs=clevs, vmin=vmin, vmax=vmax)
    meteva.base.plot_tools.plot_2d_grid_list(grd_list,cmap=cmap1,clevs = clevs1,ncol= 2)




if __name__ == "__main__":

    filename_ob = r'H:\test_data\input\mem\mode\ob\rain03\20072611.000.nc'
    filename_fo = r'H:\test_data\input\mem\mode\ec\rain03\20072608.003.nc'
    grd_ob = meteva.base.read_griddata_from_nc(filename_ob)
    grd_fo = meteva.base.read_griddata_from_nc(filename_fo)
    look_cra = operation(grd_ob,grd_fo,1,5,30)


import pandas as pd
import copy
import meteva
import datetime
import math
import numpy as np


def rain_to_01process(rain01h_ob_fos,used_coords = "time"):
    '''
    降水过程必须被完整包含在时间（时效）序列内，否则起止时间会有偏差
    :param rain01h_ob_fos:
    :param used_coords:
    :return:
    '''

    if used_coords=="time":
        step = datetime.timedelta(hours=1)
    else:
        step = 1
    rain01h_ob_fos_hap = rain01h_ob_fos.copy()
    values = rain01h_ob_fos_hap.iloc[:,6:].values
    values[values<0.1] = 0
    values[values>0] = 1
    rain01h_ob_fos_hap.iloc[:,6:] = values[...]
    rain01h_ob_fos_hap_move01 = rain01h_ob_fos_hap.copy()
    rain01h_ob_fos_hap_move01[used_coords] -= step
    rain01h_ob_fos_hap.iloc[:,6:] += 10
    rain01h_ob_fos_hap = meteva.base.max_on_level_time_dtime_id(rain01h_ob_fos_hap,rain01h_ob_fos_hap_move01,how = "outer",default=0)
    rain01h_ob_fos_hap_move01[used_coords] += 2 * step
    rain01h_ob_fos_hap = meteva.base.max_on_level_time_dtime_id(rain01h_ob_fos_hap, rain01h_ob_fos_hap_move01,how = "outer",default=0)
    values2 = rain01h_ob_fos_hap.iloc[:, 6:].values
    values2[values2>=10] -=10
    rain01h_ob_fos_hap.iloc[:, 6:] =values2[...]
    rain02h_ob_fos_hap = meteva.base.sum_of_sta(rain01h_ob_fos_hap,used_coords=[used_coords],span = 2)
    values3 = rain02h_ob_fos_hap.iloc[:,6:].values
    values3[values3>0] = 1
    rain02h_ob_fos_hap.iloc[:,6:] = values3[...]
    rain02h_ob_fos_hap_back01 = rain02h_ob_fos_hap.copy()
    rain02h_ob_fos_hap_back01[used_coords] -= step
    rain02h_ob_fos_hap_sprate02 =meteva.base.min_on_level_time_dtime_id(rain02h_ob_fos_hap,rain02h_ob_fos_hap_back01,how = "outer",default=0)

    rain02h_ob_fos_hap_sprate02.sort_values(by = ["level","time","dtime","id"],inplace=True)
    return  rain02h_ob_fos_hap_sprate02

def rain_process_start(rain01h_ob_fos,used_coords = "time"):
    '''
    在时间维度，统计降水的开始时间
    :param rain01h_ob_fos:
    :return:
    '''
    rain02h_ob_fos_hap_sprate02 = rain_to_01process(rain01h_ob_fos,used_coords=used_coords)
    delta = meteva.base.change(rain02h_ob_fos_hap_sprate02,used_coords=used_coords,delta=1)
    # if used_coords =="time":
    #     delta[used_coords] -= datetime.timedelta(hours=1)
    # else:
    #     delta[used_coords] -= 1
    values = delta.iloc[:, 6:].values
    values[values <=0] = 0
    rain_start_01 = delta.copy()
    rain_start_01.iloc[:, 6:] = values[...]
    return rain_start_01


def rain_process_end(rain01h_ob_fos,used_coords = "time"):
    '''
    在时间维度，统计降水的开始时间
    :param rain01h_ob_fos:
    :return:
    '''
    rain02h_ob_fos_hap_sprate02 = rain_to_01process(rain01h_ob_fos,used_coords=used_coords)
    delta = meteva.base.change(rain02h_ob_fos_hap_sprate02,used_coords=used_coords,delta=1)
    if used_coords =="time":
        delta[used_coords] -= datetime.timedelta(hours=1)
    else:
        delta[used_coords] -= 1

    values = delta.iloc[:, 6:].values
    values[values >=0] = 0
    values *= -1
    rain_end_01 = delta.copy()
    rain_end_01.iloc[:, 6:] = values[...]
    return rain_end_01


def rain_process_lenght(rain01h_ob_fos,used_coords = "time",record_on = "end"):
    rain02h_ob_fos_hap_sprate02 = rain_to_01process(rain01h_ob_fos,used_coords=used_coords)
    rain02h_ob_fos_hap_sprate02_m = rain02h_ob_fos_hap_sprate02.copy()
    lenght = rain02h_ob_fos_hap_sprate02.copy()


    if record_on =="end":
        move_step = 1
    else:
        move_step  = -1

    if used_coords =="time":
        max_move = 72
        step =  datetime.timedelta(hours=move_step)
    else:
        dtime_list = list(set(rain02h_ob_fos_hap_sprate02["dtime"].values))
        dtime_max = np.max(np.array(dtime_list))
        max_move = min(dtime_max, 72)
        step = move_step

    for dh in range(max_move):
        rain02h_ob_fos_hap_sprate02_m[used_coords] += step
        rain02h_ob_fos_hap_sprate02_m = meteva.base.mutiply_on_level_time_dtime_id(rain02h_ob_fos_hap_sprate02_m,
                                                                                   rain02h_ob_fos_hap_sprate02)
        lenght = meteva.base.add_on_level_time_dtime_id(lenght, rain02h_ob_fos_hap_sprate02_m,how="left",default=0)
    delta = meteva.base.change(rain02h_ob_fos_hap_sprate02, used_coords=used_coords, delta=1)

    if record_on=="end":
        if used_coords == "time":
            delta[used_coords] -= datetime.timedelta(hours=1)
        else:
            delta[used_coords] -= 1

    values = -move_step *delta.iloc[:, 6:].values
    values[values <=0] = 0
    rain_end_01 = delta.copy()
    rain_end_01.iloc[:, 6:] = values[...]
    lenght = meteva.base.mutiply_on_level_time_dtime_id(lenght,rain_end_01,how="left",default=0)
    return lenght


def rain_process_peak(rain01h_ob_fos,used_coords = "time"):

    process01 = rain_to_01process(rain01h_ob_fos,used_coords=used_coords)
    rain01h_ob_fos_sprate02 = meteva.base.mutiply_on_level_time_dtime_id(rain01h_ob_fos,process01,how="right",default=0)
    values0 = rain01h_ob_fos_sprate02.iloc[:,6:].values
    if(len(values0.shape) ==2):
        rand = 0.001*np.random.rand(values0.shape[0],values0.shape[1])
    else:
        rand = 0.001 * np.random.rand(values0.shape[0])
    values0[values0>=0.1] += rand[values0>=0.1]  # 在原值上增加微量的随机增量，以避免一个过程中完全相同的两个峰值导致峰值时间的频次不合理。

    rain01h_ob_fos_sprate02.iloc[:,6:]  = values0[...]
    rain01h_ob_fos_move = rain01h_ob_fos_sprate02.copy()
    max_value = rain01h_ob_fos_sprate02.copy()
    if used_coords =="time":
        max_move = 72
        step =  datetime.timedelta(hours=1)
    else:
        dtime_list = list(set(process01["dtime"].values))
        dtime_max = np.max(np.array(dtime_list))
        max_move = min(dtime_max, 72)
        step = 1

    for dh in range(1,max_move):
        rain01h_ob_fos_move[used_coords] += step
        rain01h_ob_fos_move = meteva.base.mutiply_on_level_time_dtime_id(rain01h_ob_fos_move,
                                                                                   process01,how="right",default=0)
        max_value = meteva.base.max_on_level_time_dtime_id(max_value, rain01h_ob_fos_move)
        print("已完成"+ str(int(100*dh/(2*max_move+1)))+"%")

    rain01h_ob_fos_move = rain01h_ob_fos_sprate02.copy()
    for dh in range(1,max_move):
        rain01h_ob_fos_move[used_coords] -= step
        rain01h_ob_fos_move = meteva.base.mutiply_on_level_time_dtime_id(rain01h_ob_fos_move,
                                                                          process01,how="right",default=0)
        max_value = meteva.base.max_on_level_time_dtime_id(max_value, rain01h_ob_fos_move)
        print("已完成" + str(int(100*(dh+max_move) / (2 * max_move + 1)))+"%")
    values = rain01h_ob_fos_sprate02.iloc[:, 6:].values
    values[values==0] = -1
    rain01h_ob_fos_sprate02.iloc[:, 6:] = values[...]
    peak = meteva.base.minus_on_level_time_dtime_id(rain01h_ob_fos_sprate02,max_value,how="left",default=0)
    values2 = peak.iloc[:, 6:].values
    values2[values2<0] =-1
    values2 +=1
    peak.iloc[:, 6:] = values2[...]

    peak_time = rain01h_ob_fos.copy()
    peak_time.iloc[:,6:] = 0
    peak_time = meteva.base.max_on_level_time_dtime_id(peak_time,peak,how = "left",default=0)
    peak_value = meteva.base.mutiply_on_level_time_dtime_id(rain01h_ob_fos,peak_time)

    return peak_time,peak_value


def is_max_of_oneday(sta,used_coords = "time",start_hour_of_one_day = 0,keep_zeros = True):
    '''
    判断一个数据样本是否为一天中的最大值，是则记为1，否则记为0
    :param sta:
    :param used_coords:
    :param start_hour_of_one_day:
    :return:
    '''
    pass

    sta_max = meteva.base.max_of_sta(sta,used_coords=used_coords,span = 24)
    sta_max_last_time = meteva.base.sele_by_para(sta_max,ob_hour=start_hour_of_one_day)
    if len(sta_max_last_time.index) ==0:
        print("没有最大值观测时间和"+str(start_hour_of_one_day).zfill(2)+"时对应的数据样本")
    sta_max_list = [sta_max_last_time]
    for dh in range(1,24):
        sta_max_m = copy.deepcopy(sta_max_last_time)
        if used_coords =="time":
            sta_max_m["time"] -= datetime.timedelta(hours=dh)
        elif used_coords=="dtime":
            sta_max_m["dtime"] -= dh
        else:
            print("参数used_coords 只支持 time 和dtime两个选项")
            return
        sta_max_list.append(sta_max_m)
    sta_max_all = meteva.base.concat(sta_max_list)
    sta_minus = meteva.base.minus_on_level_time_dtime_id(sta,sta_max_all,how="inner")
    values = sta_minus.values[:,6:]
    values[values==0] = 1
    values[values<0] = 0
    sta_minus.iloc[:, 6:] =values[:,:]
    if not keep_zeros:
        sta0 = copy.deepcopy(sta)
        values = sta0.values[:, 6:]
        values[values != 0] = 1
        sta0.iloc[:,6:] = values
        sta_minus = meteva.base.mutiply_on_level_time_dtime_id(sta_minus,sta0,how="inner")

    return sta_minus



def is_min_of_oneday(sta,used_coords = "time",start_hour_of_one_day = 0):
    '''
    判断一个数据样本是否为一天中的最小值，是则记为1，否则记为0
    :param sta:
    :param used_coords:
    :param start_hour_of_one_day:
    :return:
    '''
    pass

    sta_min = meteva.base.min_of_sta(sta,used_coords=used_coords,span = 24)
    sta_min_last_time = meteva.base.sele_by_para(sta_min,ob_hour=start_hour_of_one_day)
    if len(sta_min_last_time.index) ==0:
        print("没有最大值观测时间和"+str(start_hour_of_one_day).zfill(2)+"时对应的数据样本")
        return
    sta_min_list = [sta_min_last_time]
    for dh in range(1,24):
        sta_min_m = copy.deepcopy(sta_min_last_time)
        if used_coords =="time":
            sta_min_m["time"] -= datetime.timedelta(hours=dh)
        elif used_coords=="dtime":
            sta_min_m["dtime"] -= dh
        else:
            print("参数used_coords 只支持 time 和dtime两个选项")
            return
        sta_min_list.append(sta_min_m)
    sta_min_all = meteva.base.concat(sta_min_list)
    sta_minus = meteva.base.minus_on_level_time_dtime_id(sta_min_all,sta,how="inner")
    values = sta_minus.values[:,6:]
    values[values==0] = 1
    values[values<0] = 0
    sta_minus.iloc[:, 6:] =values[:,:]


    return sta_minus


import meteva
import datetime
import time
import os
import pandas as pd

def read_stadata_from_gds_file_or_service(binary_root,ip,port,filename,station,gds_filename_list = None):
    filename1 = binary_root + "/" + filename
    if os.path.exists(filename1):
        #print("从本地读取数据")
        sta = meteva.base.read_stadata_from_gds_griddata_file(filename1,station)
        return sta
    else:
        #print("从gds读取数据")
        dir, filename2 = os.path.split(filename)
        if gds_filename_list is None:
            gds_filename_list = meteva.base.path_tools.get_gds_file_list_in_one_dir(ip, port, dir)
        if filename2 in gds_filename_list:
            sta = meteva.base.read_stadata_from_gds_griddata(ip, port, dir + "/" + filename2, station)
            return sta
        else:
            #print(filename + "文件不存在")
            return None


def gds_ob_multi_time_fo(para):
    now = datetime.datetime.now()
    now = datetime.datetime(now.year, now.month, now.day, now.hour, 0)
    for file in para["ip_port_file"]:
        if os.path.exists(file):
            ip, port = meteva.base.read_gds_ip_port(file)
            break
    df = pd.DataFrame({"id": para["station_id_list"],
                       "lon":para["station_lon_list"],
                       "lat":para["station_lat_list"]})
    station = meteva.base.sta_data(df)
    station["data0"] = 1e10
    #print(station)
    id_list = para["station_id_list"]
    fo_time_list = []
    for i in range(para["max_dh"] + 12):
        time0 = now - datetime.timedelta(hours=i)
        if time0.hour in para["update_hours"]:
            fo_time_list.append(time0)
    #time0 = now - datetime.timedelta(hours =24)
    #path = meteva.base.get_path(para["ob_dir"], time0)
    #meteva.base.print_gds_file_values_names(ip, port, path)

    if para["value_type"] == "风":
        dir_u, file_u = os.path.split(para["fo_grd_dir_u"])
        file_list_u = meteva.base.path_tools.get_gds_file_list_in_one_dir(ip, port, dir_u)
        dir_v, file_v = os.path.split(para["fo_grd_dir_v"])
        file_list_v = meteva.base.path_tools.get_gds_file_list_in_one_dir(ip, port, dir_v)

        # 读取预报数据
        sta_fo_all = None
        ob_time_list = []
        print("开始加载预报数据")
        begin = time.time()
        for i in range(len(fo_time_list)):
            time0 = fo_time_list[i]
            for dh in range(para["max_dh"]):
                filename = meteva.base.get_path(file_u, time0, dh)
                if not filename in file_list_u: continue
                sta_u = meteva.base.read_stadata_from_gds_griddata(ip, port, dir_u + "/" + filename, station)
                if (sta_u is None): continue
                meteva.base.set_stadata_coords(sta_u, time=time0, dtime=dh)
                filename = meteva.base.get_path(file_v, time0, dh)
                if not filename in file_list_v: continue
                sta_v = meteva.base.read_stadata_from_gds_griddata(ip, port, dir_v + "/" + filename, station)
                if sta_v is None: continue
                meteva.base.set_stadata_coords(sta_v, time=time0, dtime=dh)
                sta_uv = meteva.base.combine_on_all_coords(sta_u, sta_v)
                meteva.base.set_stadata_names(sta_uv, data_name_list=["u" + para["title"], "v" + para["title"]])
                sta_fo_all = meteva.base.combine_join(sta_fo_all, sta_uv)
                ob_time = time0 + datetime.timedelta(hours=dh)
                ob_time_list.append(ob_time)
                #print(filename)
            if(i>1):
                time_end  = time.time()
                time_left = int((time_end - begin) * (len(fo_time_list) - i - 1) / (i+1 )) + 1
                print("载入预报场还需" + str(time_left) + "秒")

        ob_time_list = set(ob_time_list)
        ob_time_list1 = []
        for ob_time in ob_time_list:
            if ob_time <= now:
                ob_time_list1.append(ob_time)
        ob_time_list1.sort()
        print("开始加载观测数据")
        begin = time.time()
        sta_ob_s_all = None
        sta_ob_d_all = None
        for i in range(len(ob_time_list1)):
            time0 = ob_time_list1[i]
            path = meteva.base.get_path(para["ob_dir"], time0)
            element_id = meteva.base.gds_element_id_dict[para["ob_s_name"]]
            sta_s = meteva.base.read_stadata_from_gds(ip, port, filename=path, element_id=element_id,station=station)
            if sta_s is None: continue
            sta_ob_s_all = meteva.base.combine_join(sta_ob_s_all, sta_s)
            element_id = meteva.base.gds_element_id_dict[para["ob_d_name"]]
            sta_d = meteva.base.read_stadata_from_gds(ip, port, filename=path, element_id=element_id,station=station)
            if sta_d is None: continue
            sta_ob_d_all = meteva.base.combine_join(sta_ob_d_all, sta_d)
            if i >1:
                time_end = time.time()
                time_left = int((time_end - begin) * (len(ob_time_list1) - i - 1) / (i + 1)) + 1
                print("载入观测数据还需" + str(time_left) + "秒")

        sta_ob_s_all = meteva.base.between_value_range(sta_ob_s_all, -1e10, 1e9)
        sta_ob_d_all = meteva.base.between_value_range(sta_ob_d_all, -1e10, 1e9)
        sta_ob_all = meteva.base.diag.speed_angle_to_wind(sta_ob_s_all, sta_ob_d_all)
        for id in id_list:
            output_path = para["output_dir"]
            if (output_path is not None):
                output_path = meteva.base.get_path(output_path, now)
                output_path = output_path + "/" + str(id) + ".png"
            sta_ob = meteva.base.in_id_list(sta_ob_all, [id])
            sta_fo = meteva.base.in_id_list(sta_fo_all, [id])
            meteva.nmc_vf_product.wind_ob_and_multi_time_fo(sta_ob, sta_fo, pic_path=output_path, max_dh=para["max_dh"],
                                                     plot_error=True)
    else:
        dir, file = os.path.split(para["fo_grd_dir"])
        file_list = meteva.base.path_tools.get_gds_file_list_in_one_dir(ip, port, dir)
        # 读取预报数据
        print("开始加载预报数据")
        begin = time.time()
        sta_fo_all = None
        ob_time_list = []
        for i in range(len(fo_time_list)):
            time0 = fo_time_list[i]
            for dh in range(para["max_dh"]):
                filename = meteva.base.get_path(file, time0, dh)
                sta_u = read_stadata_from_gds_file_or_service(para["local_root"],ip,port,dir + "/" + filename, station,file_list)
                #if not filename in file_list: continue
                #sta_u = meteva.base.read_stadata_from_gds_griddata(ip, port, dir + "/" + filename, station)
                if (sta_u is None): continue
                ob_time = time0 + datetime.timedelta(hours=dh)
                ob_time_list.append(ob_time)
                meteva.base.set_stadata_coords(sta_u, time=time0, dtime=dh)
                meteva.base.set_stadata_names(sta_u, data_name_list=[para["title"]])
                sta_fo_all = meteva.base.combine_join(sta_fo_all, sta_u)
            time_end  = time.time()
            time_left = int((time_end - begin) * (len(fo_time_list) - i - 1) / (i+1 )) + 1
            print("载入预报场还需" + str(time_left) + "秒")

        ob_time_list = set(ob_time_list)
        ob_time_list1 = []
        for ob_time in ob_time_list:
            if ob_time <= now:
                ob_time_list1.append(ob_time)
        ob_time_list1.sort()
        print("开始加载观测数据")
        begin = time.time()
        sta_ob_all = None
        for i in range(len(ob_time_list1)):
            time0 = ob_time_list1[i]
            path = meteva.base.get_path(para["ob_dir"], time0)
            sta_s = meteva.base.read_stadata_from_gds(ip, port, filename=path,
                                              element_id=meteva.base.gds_element_id_dict[para["ob_name"]],
                                                                       station= station)
            if sta_s is None: continue
            sta_ob_all = meteva.base.combine_join(sta_ob_all, sta_s)
            if i > 1:
                time_end = time.time()
                time_left = int((time_end - begin) * (len(ob_time_list1) - i - 1) / (i + 1)) + 1
                print("载入观测数据还需" + str(time_left) + "秒")

        sta_ob_all = meteva.base.between_value_range(sta_ob_all,-1e10,1e9)

        for id in id_list:
            output_path = para["output_dir"]
            if (output_path is not None):
                output_path = meteva.base.get_path(output_path, now)
                output_path = output_path + "/" + str(id) + ".png"
            sta_ob = meteva.base.in_id_list(sta_ob_all, [id])
            sta_fo = meteva.base.in_id_list(sta_fo_all, [id])
            if para["value_type"] == "温度":
                meteva.nmc_vf_product.temp_ob_and_multi_time_fo(sta_ob, sta_fo, pic_path=output_path, max_dh=para["max_dh"],
                                                         plot_error=True)
            elif para["value_type"] == "降水1h":
                pass
            elif para["value_type"] == "能见度":
                meteva.nmc_vf_product.vis_ob_and_multi_time_fo(sta_ob, sta_fo, pic_path=output_path, max_dh=para["max_dh"],
                                                         plot_error=True)
            elif para["value_type"] == "相对湿度":
                meteva.nmc_vf_product.rh_ob_and_multi_time_fo(sta_ob, sta_fo, pic_path=output_path,
                                                                     max_dh=para["max_dh"],
                                                                     plot_error=True)


import numpy as np
import pandas as pd


from meteva.method.space.baddeley_binary_image_metric.lib.as_unitname import as_unitname


def im(mat, xcol=None, yrow=None, xrange=None, yrange=None, unitname=None):
    typ = type(mat)
    if typ == "double":
        typ = "real"
    miss_xcol = xcol is None
    miss_yrow = yrow is None
    xcol = range(0, mat.shape[0])
    yrow = range(0, mat.shape[1])
    if mat.shape is not None:
        nr = mat.shape[1]
        nc = mat.shape[0]
        if np.isnan(nc):
            nc = 1
            nr = len(mat)
            if miss_xcol:
                xcol = range(0, nc)
        if len(xcol) != nc:
            raise Exception("Length of xcol does not match ncol(mat)")
        if len(yrow) != nr:
            raise Exception("Length of yrow does not match nrow(mat)")
    else:
        if miss_xcol or miss_yrow:
            raise Exception("mat is not a matrix and I can\'t guess its dimensions")
        if len(mat) != len(xcol) * len(yrow):
            return
        nc = len(xcol)
        nr = len(yrow)

    # lev = None
    # if 'categories' in mat:
    #     lev = mat.categories
    # if mat.dtype == 'category':
    #     typ = "category"
    # elif lev is not None:
    #     typ = "category"
    #     mat = pd.Categorical(mat, categories=lev)
    if type(mat) != pd.DataFrame:
        mat.reshape(nr, nc)
    if (miss_xcol or len(xcol) <= 1) and xrange is not None:
        xstep = np.diff(xrange) / nc
        xcol = np.linspace(start=xrange[0] + xstep / 2, stop=xrange[1] - xstep / 2, num=nc)
    elif len(xcol) > 1:
        xcol = np.linspace(start=min(xcol), stop=max(xcol), num=len(xcol))
        xstep = np.diff(xcol)[0]
        xrange = np.array([min(xcol), max(xcol)]) + np.array([-1, 1]) * xstep / 2
    else:
        raise Exception("Cannot determine pixel width")
    if (miss_yrow or len(yrow) <= 1) and yrange is not None:
        ystep = np.diff(yrange) / nr
        yrow = np.linspace(start=yrange[1] + ystep / 2, stop=yrange[2] - ystep / 2, num=nr)
    elif len(yrow) > 1:
        yrow = np.linspace(start=min(yrow), stop=max(yrow), num=len(yrow))
        ystep = np.diff(yrow)[0]
        yrange = np.array([min(yrow), max(yrow)]) + np.array([-1, 1]) * ystep / 2
    else:
        raise Exception("Cannot determine pixel height")
    unitname = as_unitname(unitname)
    out = {'v': mat, 'dim': [nr, nc], 'xrange': xrange, 'yrange': yrange,
           'xstep': xstep, 'ystep': ystep, 'xcol': xcol, 'yrow': yrow, 'type': typ, 'units': unitname}
    return out



import meteva
from meteva.base import *
from meteva.method import *
from meteva.product.program.fun import *
import pandas as pd
import numpy as np




def diunal_max_hour(sta_ob_and_fos0,method,s = None,g = None,gll = None,group_name_list = None,plot = None,
          vmax = 24,vmin = 0,bar_width = None,save_path = None,show = False,dpi = 300,title = "",excel_path = None,**kwargs):
    '''
    对于每个分组的样本，先统计日变化的曲线，再求日变化曲线的顶点
    :param sta_ob_and_fos0:
    :param method:
    :param s:
    :param g:
    :param gll:
    :param group_name_list:
    :param plot:
    :param vmax:
    :param vmin:
    :param bar_width:
    :param save_path:
    :param show:
    :param dpi:
    :param title:
    :param excel_path:
    :param kwargs:
    :return:
    '''
    if g=="ob_hour":
        print("日变化分析模块已经包含了按ob_hour分类统计功能，不能再设置g = ob_hour")
    if s is not None:
        if g is not None:
            if g == "last_range" or g == "last_step":
                s["drop_last"] = False
            else:
                s["drop_last"] = True
    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0, s=s)
    if (len(sta_ob_and_fos1.index) == 0):
        print("there is no data to verify")
        return

    dtimes = sta_ob_and_fos1["dtime"] * np.timedelta64(1, 'h')
    obtimes = pd.Series(0,index = sta_ob_and_fos1['time'] + dtimes)
    ob_hours_list = list(set(obtimes.index.hour))
    ob_hours_list.sort()
    ob_hour_count = len(ob_hours_list)

    method_args = {}
    plot_args = {}
    method_para_list = method.__code__.co_varnames
    plot_mehod = None
    if plot == "bar":
        plot_mehod = meteva.base.plot_tools.bar
    elif plot == "line" or plot == "plot":
        plot_mehod = meteva.base.plot_tools.plot

    plot_para_list = []
    if plot_mehod is not None:
        plot_para_list = plot_mehod.__code__.co_varnames

    for key in kwargs.keys():
        if key in method_para_list and key not in plot_para_list:
            method_args[key] = kwargs[key]
        elif key not in method_para_list and key in plot_para_list:
            plot_args[key] = kwargs[key]
        elif key in method_para_list and key in plot_para_list:
            msg = method.__name__ + " and " + plot_mehod.__name__ + " have same args:" + key
            print(msg)
            return None, msg
        else:
            if plot_mehod is None:
                msg = key + " is not args of " + method.__name__
            else:
                msg = key + " is not args of " + method.__name__ + " or " + plot_mehod.__name__
            print(msg)
            return None, msg

    #等级参数的确定
    if "grade_list" not in kwargs.keys():
        mutil_list = [meteva.method.ts_multi, meteva.method.bias_multi, meteva.method.ets_multi,
                      meteva.method.mr_multi, meteva.method.far_multi]
        if method in mutil_list:
        # 如果是多分类检验，但又没有设置分级方法，就需要从数据中获得全局的种类
            values = sta_ob_and_fos1.iloc[:,6:].flatten()
            index_list = list(set(values))
            if len(index_list) > 30:
                msg = "自动识别的样本类别超过30种，判断样本为连续型变量，grade_list不能缺省"
                print(msg)
                return None,msg
            index_list.sort()
            grade_list = []
            for i in range(len(index_list)-1):
                grade_list.append((index_list[i]+index_list[i+1])/2)
            kwargs["grade_list"] = grade_list

    if "grade_list" in kwargs.keys():
        grades = kwargs["grade_list"]
        grade_names = []
        mutil_list1 = [meteva.method.ts_multi, meteva.method.bias_multi, meteva.method.ets_multi,
                       meteva.method.mr_multi, meteva.method.far_multi,
                       meteva.method.ts_grade, meteva.method.bias_grade, meteva.method.ets_grade,
                       meteva.method.mr_grade, meteva.method.far_grade]
        mutil_list2 = [meteva.method.accuracy,meteva.method.hk,meteva.method.hss]
        if method in mutil_list1:
            grade_names = ["<" +  str(grades[0])]
            for i in range(len(grades)-1):
                grade_names.append("["+str(grades[i]) + ","+str(grades[i+1])+")")
            grade_names.append(">=" + str(grades[-1]))
        elif method in mutil_list2:
            grade_names = ["0"]
        else:
            for i in range(len(grades)):
                grade_names.append(grades[i])
    else:
        grade_names = ["0"]
    grade_num = len(grade_names)


    sta_ob_and_fos_list,gll1 = meteva.base.group(sta_ob_and_fos1,g = g,gll=gll)
    if gll1 is None:gll1 = [None]
    data_name = meteva.base.get_stadata_names(sta_ob_and_fos_list[0])
    if method.__name__.find("ob_fo")>=0:
        fo_name = data_name
    elif method.__name__.find("_uv")>=0:
        fo_name = []
        for i in range(2,len(data_name),2):
            strs = data_name[i]
            strs = strs.replace("u_","")
            fo_name.append(strs)
    else:
        ensemble_score_method = [meteva.method.cr,variance_divide_by_mse]
        if method in ensemble_score_method:
            fo_name = [""]
        elif method ==meteva.method.variance_mse:
            fo_name = ["variance","MSE"]
        elif method ==meteva.method.std_rmse:
            fo_name = ["STD","RMSE"]
        else:
            fo_name = data_name[1:]

    fo_num = len(fo_name)

    gll_valid = []
    result_list =[]
    for i in range(len(gll1)):
        result_all,hour_list = meteva.product.program.score(sta_ob_and_fos_list[i],method,g = "ob_hour",gll = ob_hours_list,**method_args)
        hours = np.array(hour_list)
        if len(hour_list) == ob_hour_count:
            gll_valid.append(gll1[i])
            result_max_index = np.argmax(result_all,axis=0)
            result_max = hours[result_max_index]
            result_list.append(result_max)

    result = np.array(result_list)
    group_num = len(gll_valid)

    if plot is not None or excel_path is not None:

        name_list_dict = {}
        if g is None:
            group_dict_name = "group_name"
        else:
            group_dict_name = g
        #设置分组名称
        if group_name_list is not None:
            if group_num == len(group_name_list):
                name_list_dict[group_dict_name] =group_name_list
            else:
                print("group_name_list参数中包含的分组名称个数和实际分组个数不匹配")
        else:
            if not isinstance(gll_valid,list):
                group_list_list1 = [gll_valid]

            if (group_dict_name == "time" or group_dict_name == "ob_time")and gll is None:
                name_list_dict[group_dict_name] = gll_valid
            else:
                name_list_dict[group_dict_name] = get_group_name(gll_valid)

        #设置成员名称

        name_list_dict["member"] = fo_name
        if fo_num==0:
            fo_num = 1
            name_list_dict["member"] = ["OBS"]
        result_plot = result.reshape((group_num,fo_num,grade_num))

        #设置等级名称
        name_list_dict["grade"] = grade_names
        keys = list(name_list_dict.keys())
        if fo_num ==1:
            if grade_num > 1:
                if group_num>1:
                    legend = keys[2]
                    axis = keys[0]
                else:
                    axis = keys[2]
                    legend = keys[1]
            else:
                axis = keys[0]
                legend = keys[1]
        else:
            if group_num == 1 :
                if grade_num >1:
                    legend = keys[1]
                    axis = keys[2]
                else:
                    legend = keys[2]
                    axis = keys[1]
            else:
                legend = keys[1]
                axis = keys[0]
        if "ylabel" not in plot_args.keys():
            plot_args["ylabel"] = "峰值时间"



        bigthan0_method = [meteva.method.ts,meteva.method.ob_fo_hr,meteva.method.ob_fo_std,meteva.method.ts_multi,
                           meteva.method.s,meteva.method.pc_of_sun_rain,meteva.method.bias_multi,meteva.method.bias,
                           meteva.method.pc,meteva.method.mr,meteva.method.far,meteva.method.tc,
                           meteva.method.roc_auc,meteva.method.r,meteva.method.sr,meteva.method.cr,meteva.method.pod,
                           meteva.method.pofd,meteva.method.mse,meteva.method.rmse,meteva.method.mae]

        if vmin is None:
            if method in bigthan0_method:
                vmin = 0
        if "hline" not in plot_args.keys():
            if method in [meteva.method.me]:
                plot_args["hline"] = 0
            elif method in [meteva.method.bias]:
                plot_args["hline"] = 1


        if plot is not None:
            if plot =="bar":
                meteva.base.plot_tools.bar(result_plot,name_list_dict,legend=legend,axis = axis,vmin =vmin,vmax = vmax,bar_width=bar_width,save_path=save_path,show=show,dpi =dpi,title = title,**plot_args)
            else:
                meteva.base.plot_tools.plot(result_plot,name_list_dict,legend=legend,axis = axis,vmin =vmin,vmax = vmax,save_path=save_path,show=show,dpi = dpi,title= title,**plot_args)
        if excel_path is not None:
            meteva.base.write_array_to_excel(result_plot,excel_path,name_list_dict,index= axis,columns=legend)

    return result,gll_valid



def diunal_max_hour_id(sta_ob_and_fos0,method,s = None,plot = "scatter",save_dir = None,save_path = None,show = False,
             add_county_line = False,map_extend= None,print_max=0,print_min=0,dpi = 300,title = None,sort_by = None,
             **kwargs):
    '''

    :param sta_ob_and_fos0:
    :param method:
    :param s:
    :param plot:
    :param save_dir:
    :param save_path:
    :param show:
    :param add_county_line:
    :param map_extend:
    :param print_max:
    :param print_min:
    :param dpi:
    :param title:
    :param sort_by:
    :param kwargs:
    :return:
    '''
    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0, s=s)
    diunal_max_hour(sta_ob_and_fos1,method)
    data_names = meteva.base.get_stadata_names(sta_ob_and_fos1)
    ensemble_score_method = [meteva.method.cr, variance_divide_by_mse]

    # get method_args and plot_args
    method_args = {}
    plot_args = {}
    method_para_list = method.__code__.co_varnames
    plot_mehod = None
    if plot == "scatter":
        plot_mehod = meteva.base.plot_tools.scatter_sta

    plot_para_list = []
    if plot_mehod is not None:
        plot_para_list = plot_mehod.__code__.co_varnames


    for key in kwargs.keys():
        if key in method_para_list and key not in plot_para_list:
            method_args[key] = kwargs[key]
        elif key not in method_para_list and key  in plot_para_list:
            plot_args[key] = kwargs[key]
        elif key  in method_para_list and key  in plot_para_list:
            print(method.__name__ + " and " + plot_mehod.__name__ + " have same args:" + key)
            return
        else:
            print(key + " is not args of " + method.__name__ + " or " + plot_mehod.__name__)
            return

    #plot_args["subplot"] = "member"

    if method.__name__.find("ob_fo") >= 0:
        fo_name = data_names

    elif method.__name__.find("_uv") >= 0:
        fo_name = []
        for i in range(2, len(data_names), 2):
            strs = data_names[i]
            strs = strs.replace("u_", "")
            fo_name.append(strs)

    elif method.__name__ == "sample_count":
        fo_name = [data_names[0]]
    elif method in ensemble_score_method:
        fo_name = [""]
    elif method == meteva.method.variance_mse:
        fo_name = ["variance", "MSE"]
    elif method == meteva.method.std_rmse:
        fo_name = ["STD", "RMSE"]
    else:
        fo_name = data_names[1:]
    fo_num = len(fo_name)
    if "grade_list" in kwargs.keys():
        grades = kwargs["grade_list"]
        grade_names = []
        mutil_list1 = [meteva.method.ts_multi, meteva.method.bias_multi, meteva.method.ets_multi,
                       meteva.method.mr_multi, meteva.method.far_multi,
                       meteva.method.ts_grade, meteva.method.bias_grade, meteva.method.ets_grade,
                       meteva.method.mr_grade, meteva.method.far_grade]
        mutil_list2 = [meteva.method.accuracy,meteva.method.hk,meteva.method.hss]
        if method in mutil_list1:
            grade_names = ["<" +  str(grades[0])]
            for i in range(len(grades)-1):
                grade_names.append("["+str(grades[i]) + ","+str(grades[i+1])+")")
            grade_names.append(">=" + str(grades[-1]))
        elif method in mutil_list2:
            grade_names = ["0"]
        else:
            for i in range(len(grades)):
                grade_names.append(grades[i])
    else:
        grade_names = ["0"]
    grade_num = len(grade_names)

    if save_path is not None:
        if isinstance(save_path, str):
            save_path = [save_path]


    result, id_list = diunal_max_hour(sta_ob_and_fos1, method, g="id", **method_args,plot = None)
    station = sta_ob_and_fos1.drop_duplicates(['id'], inplace=False)
    station.iloc[:, 1] = station.iloc[0, 1]
    station.iloc[:, 2] = station.iloc[0, 2]
    station.iloc[:, 0] = station.iloc[0, 0]
    station1 = meteva.base.in_id_list(station, id_list)
    id_s = pd.Series(id_list)
    id_s.name = "id"
    sta_merge = pd.merge(id_s, station1, on='id', how='left')
    sta_merge = meteva.base.sta_data(sta_merge)
    coord_names = meteva.base.get_coord_names()

    if len(result.shape) == 1:
        # 没有等级，只有一个预报成员
        coord_names.append(data_names[-1])
        sta_result = sta_merge.loc[:, coord_names]
        sta_result.iloc[:, -1] = result[:]
        sta_result = [sta_result]
    elif len(result.shape) == 2:
        if len(fo_name) > 1:
            # 没有等级，但有多个预报成员
            if method.__name__.find("_uv") >= 0:
                # member_num = result.shape[1]
                # coord_names.extend(fo_name)
                sta_result = sta_merge.loc[:, coord_names]
                for ff in range(len(fo_name)):
                    sta_result[fo_name[ff]] = result[:, ff]
                sta_result = [sta_result]
            elif method == meteva.method.variance_mse:
                member_num = result.shape[1]
                # coord_names.extend(fo_name)
                sta_result = sta_merge.loc[:, coord_names]
                sta_result["variance"] = result[:, 0]
                sta_result["MSE"] = result[:, 1]
                sta_result = [sta_result]
            elif method == meteva.method.std_rmse:
                member_num = result.shape[1]
                # coord_names.extend(fo_name)
                sta_result = sta_merge.loc[:, coord_names]
                sta_result["STD"] = result[:, 0]
                sta_result["RMSE"] = result[:, 1]
                sta_result = [sta_result]
            else:
                member_num = result.shape[1]
                coord_names.extend(fo_name)
                sta_result = sta_merge.loc[:, coord_names]
                sta_result.iloc[:, -member_num:] = result[:, :]
                sta_result = [sta_result]
        else:
            # 有多个等级，但只有一个预报成员
            sta_result = []
            coord_names.append(data_names[-1])
            for i in range(result.shape[1]):
                sta_result1 = sta_merge.loc[:, coord_names]
                sta_result1.iloc[:, -1] = result[:, i]
                sta_result.append(sta_result1)
    else:
        # 有多个等级，同时有多个预报成员
        sta_result = []
        coord_names.extend(fo_name)
        member_num = result.shape[1]
        for i in range(result.shape[2]):
            sta_result1 = sta_merge.loc[:, coord_names]
            sta_result1.iloc[:, -member_num:] = result[:, :, i]
            sta_result.append(sta_result1)
    for i in range(len(sta_result)):
        sta_result1 = sta_result[i]
        if "var_name" in sta_ob_and_fos0.attrs.keys():
            sta_result1.attrs["var_name"] = sta_ob_and_fos0.attrs["var_name"]
        else:
            sta_result1.attrs["var_name"] = ""
        sta_result1.attrs["data_source"] = "meteva." + method.__name__
        if plot == "scatter" or plot == "micaps":
            if isinstance(title, list):
                title1_list = title[i: (i + 1) * fo_num]
            else:
                title1_list = []
                for ii in range(fo_num):
                    if title is not None:
                        title1 = meteva.product.program.get_title_from_dict(title, s, None, None,
                                                                            fo_name[ii])
                    else:
                        title1 = meteva.product.program.get_title_from_dict("", s, None,
                                                                           None, fo_name[ii])
                    if grade_num > 1:
                        title1 += "(grade_" + str(grade_names[i]) + ")"
                    title1_list.append(title1)

            if plot == "scatter":
                save_path1 = None
                if save_path is None:
                    if save_dir is None:
                        show = True
                    else:
                        save_path1 = []
                        for i in range(len(title1_list)):
                            fileName = title1_list[i].replace("\n", "").replace(":", "")
                            save_path1.append(save_dir + "/" + fileName + ".png")
                else:
                    save_path1 = save_path[i]

                meteva.base.tool.plot_tools.scatter_sta(sta_result1, save_path=save_path1, show=show,
                                                        title=title1_list, print_max=print_max,
                                                        print_min=print_min
                                                        , add_county_line=add_county_line,
                                                        map_extend=map_extend, dpi=dpi, **plot_args,cmap="hour")
            if plot == "micaps":
                meteva.base.put_stadata_to_micaps(sta_result1, layer_description=title1_list)

    if len(sta_result) == 1:
        sta_result = sta_result[0]
    if sort_by is not None:
        sta_result.sort_values(by=sort_by, axis=0, ascending=False, inplace=True)

    return sta_result, None





# 采用以下形式将所有.py 模块导入到fss 目录下
from . import fss
from .fss import *
#
#
# # FSS
# def FSS(Ob, Fo, window_sizes_list=[3], threshold_list=[50], Masker=None):
#     '''
#     :param Ob: 实况数据 2维的numpy
#     :param Fo: 实况数据 2维的numpy
#     :param window_sizes_list: 卷积窗口宽度的列表，以格点数为单位
#     :param threshold_list:  事件发生的阈值
#     :param Masker:  2维的numpy检验的关注区域，在Masker网格值取值为0或1，函数只对网格值等于1的区域的数据进行计算。
#     :return:
#     '''
#     shape = Ob.shape
#     nw = len(window_sizes_list)
#     # print(nw)
#     nt = len(threshold_list)
#     fss = np.zeros((nw, nt))
#     for i in range(nw):
#         kernel = np.ones((nw, nw))
#         # print(kernel)
#         ws = np.sum(kernel)
#         if Masker is not None:
#             masker_sum = np.convolve(Masker, kernel, mode="same") + 1e-10
#         else:
#             masker_sum = np.ones(shape) * ws + 1e-10
#         for j in range(nt):
#             ob_hap = np.zeros(shape)
#             ob_hap[Ob > threshold_list[j]] = 1
#             fo_hap = np.zeros(shape)
#             fo_hap[Fo > threshold_list[j]] = 1
#             ob_hap_sum = np.convolve(ob_hap, kernel, mode="same")
#             fo_hap_sum = np.convolve(fo_hap, kernel, mode="same")
#             ob_hap_p = ob_hap_sum / masker_sum
#             fo_hap_p = fo_hap_sum / masker_sum
#             a1 = np.sum(np.power(ob_hap_p - fo_hap_p, 2))
#             a2 = np.sum(np.power(ob_hap_p, 2)) + np.sum(np.power(fo_hap_p, 2))
#             fss[i, j] = 1 - a1 / (a2 + 1e-10)
#     return fss
#



from . import aaft2d
from . import ampstats
from . import cbind
from . import datagrabber_spatialVx
from . import distfun
from . import fft2d
from . import im
from . import locmeasures2d_prep
from . import locperf
from . import locperfer
from . import mae
from . import solutionset
from . import surrogater2d
from . import util
from . import zapsmall



import os
import numpy as np
import meteva



def tase_angle(d_ob, d_fo,s_ob = None,s_fo = None, ignore_breeze = False):
    '''
    计算平均误差、平均绝对误差、均方误差、均方根误差的中间结果
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 一维numpy数组，其内容依次为总样本数、误差总和、绝对误差总和、误差平方总和
    '''

    tase_list = []
    Fo_shape = d_fo.shape
    Ob_shape = d_ob.shape

    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = d_fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape

    new_Fo_s = None
    if s_fo is not None:
        new_Fo_s = s_fo.reshape(new_Fo_shape)

    ob0 = d_ob
    for line in range(new_Fo_shape[0]):
        total_count = ob0.size
        delta = new_Fo[line, :] - ob0
        delta[delta>180] = delta[delta>180] - 360
        delta[delta<-180] = 360 + delta[delta<-180]
        if ignore_breeze:
            if new_Fo_s is None or s_ob is None:
                print("if need to ignore breeze, s_fo and s_ob must not be None")
            fo_s = new_Fo_s[line,:]
            index = np.where((s_ob < 5.5)&(fo_s < 5.5))
            delta[index] = 0

        e_sum = np.sum(delta)
        ae_sum = np.sum(np.abs(delta))
        se_sum = np.sum(np.square(delta))
        tase_list.append(np.array([total_count, e_sum, ae_sum, se_sum]))
    tase_np = np.array(tase_list)
    shape = list(Fo_shape[:ind])
    shape.append(4)

    tase_array = tase_np.reshape(shape)
    return tase_array




def tase_angle_uv(u_ob, u_fo, v_ob, v_fo,ignore_breeze = False):
    '''
    计算平均误差、平均绝对误差、均方误差、均方根误差的中间结果
    -----------------------------
    :param Ob: 实况数据  任意维numpy数组
    :param Fo: 预测数据 任意维numpy数组,Fo.shape 和Ob.shape一致
    :return: 一维numpy数组，其内容依次为总样本数、误差总和、绝对误差总和、误差平方总和
    '''

    s_ob, d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob, v_ob)
    s_fo, d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo, v_fo)

    tase_array = tase_angle(d_ob, d_fo,s_ob = s_ob,s_fo = s_fo, ignore_breeze = ignore_breeze)
    return tase_array


def me_angle(d_ob,d_fo,s_ob = None,s_fo = None, ignore_breeze = False):
    '''

    :param d_ob:
    :param d_fo:
    :return:
    '''
    tase_angle_array = tase_angle(d_ob,d_fo,s_ob=s_ob,s_fo = s_fo,ignore_breeze = ignore_breeze)
    mean_error = meteva.method.me_tase(tase_angle_array)
    return mean_error


def mae_angle(d_ob, d_fo,s_ob = None,s_fo = None, ignore_breeze = False):
    '''

    :param d_ob:
    :param d_fo:
    :return:
    '''
    tase_angle_array =  tase_angle(d_ob,d_fo,s_ob=s_ob,s_fo = s_fo,ignore_breeze = ignore_breeze)
    mean_abs_error = meteva.method.mae_tase(tase_angle_array)
    return mean_abs_error

def rmse_angle(d_ob, d_fo,s_ob = None,s_fo = None, ignore_breeze = False):
    '''

    :param d_ob:
    :param d_fo:
    :return:
    '''
    tase_angle_array =  tase_angle(d_ob,d_fo,s_ob=s_ob,s_fo = s_fo,ignore_breeze = ignore_breeze)
    root_mean_sqrt_error = meteva.method.rmse_tase(tase_angle_array)
    return root_mean_sqrt_error


def me_angle_uv(u_ob, u_fo, v_ob, v_fo,ignore_breeze = False):
    '''
    风速预报评分。
    基于原始u，v风（m/s)数组，计算风速预报评分。计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当一个样本的预报和观测风速等级正好相同时，得1分，等级差1级，得0.6分，等级差2级得0.4分,否则不得分。风速评分等于所有样本得分的平均。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :return: 风速预报评分，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob, d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob, v_ob)
    s_fo, d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo, v_fo)
    mean_error = me_angle(d_ob,d_fo,s_ob = s_ob,s_fo = s_fo,ignore_breeze = ignore_breeze)
    return mean_error

def mae_angle_uv(u_ob, u_fo, v_ob, v_fo,ignore_breeze = False):
    '''
    风速预报评分。
    基于原始u，v风（m/s)数组，计算风速预报评分。计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当一个样本的预报和观测风速等级正好相同时，得1分，等级差1级，得0.6分，等级差2级得0.4分,否则不得分。风速评分等于所有样本得分的平均。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :return: 风速预报评分，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob, d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob, v_ob)
    s_fo, d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo, v_fo)
    mean_abs_error = mae_angle(d_ob,d_fo,s_ob = s_ob,s_fo = s_fo,ignore_breeze = ignore_breeze)
    return mean_abs_error


def rmse_angle_uv(u_ob, u_fo, v_ob, v_fo,ignore_breeze = False):
    '''
    风速预报评分。
    基于原始u，v风（m/s)数组，计算风速预报评分。计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当一个样本的预报和观测风速等级正好相同时，得1分，等级差1级，得0.6分，等级差2级得0.4分,否则不得分。风速评分等于所有样本得分的平均。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :return: 风速预报评分，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob, d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob, v_ob)
    s_fo, d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo, v_fo)
    root_mean_abs_error = rmse_angle(d_ob,d_fo,s_ob = s_ob,s_fo = s_fo,ignore_breeze = ignore_breeze)
    return root_mean_abs_error


def scd_nas(nasd_array):
    '''
    基于中间结果计算风向预报评分
    :param nasd_array: 输入nasd函数统计得到的样本数、风向正确样本数、风向评分（分子部分）
    :return: 返回平均的风向预报评分
    '''
    total = nasd_array[...,0]
    sc = nasd_array[...,2]
    scd = sc/total
    return scd

def acd_nas(nasd_array,unit = 1):
    '''
    基于中间结果计算风向预报准确率，
    :param nasd_array: 输入nasd函数统计得到的样本数、风向正确样本数、风向评分（分子部分）
    :return: 返回平均的风向预报准确率
    '''
    total = nasd_array[...,0]
    sc = nasd_array[...,1]
    acd = sc/total
    if unit =="%":
        acd *= 100
    return acd

def nas_d(d_ob,d_fo,s_ob = None,s_fo = None, ignore_breeze = False):
    '''
    计算风向预报准确率和评分的中间量。
    将输入的观测和预报风向（0-360度），转换成计算风向预报准确率，风向预报评分所需要的中间量
    ，计算的第一步是将预报和观测的风向都转换成8个离散的方位角。
    :param d_ob:观测的风向，numpy数组
    :param s_ob:观测的风速，numpy数组，shape和d_ob完全一致
    :param d_fo:预报的风向，numpy数组，shape可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param s_fo:预报的风速，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param ignore_breeze ： 该参数为True时，若预报和观测的风速都小于等于3级，则认为风向是正确的。
    :return:根据预报和观测的风向数据，统计得到的样本数、风向正确样本数、风向评分（分子部分），d_fo和d_ob的shape一致，则说明只有一家预报，此时返回
             一维的数据，size= 3。 如果d_fo和d_ob的shape不一致，则说明有多个预报，则返回的数据shape = （预报成员数，3）
    '''
    Ob_shape = d_ob.shape
    Fo_shape = d_fo.shape
    nasd_list = []
    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = d_fo.reshape(new_Fo_shape)
    new_Fo_s = None
    if s_fo is not None:
        new_Fo_s = s_fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape
    ob0 = meteva.base.tool.math_tools.tran_direction_to_8angle(d_ob)
    for line in range(new_Fo_shape[0]):
        fo = meteva.base.tool.math_tools.tran_direction_to_8angle(new_Fo[line, :])
        if ignore_breeze:
            if new_Fo_s is None or s_ob is None:
                print("if need to ignore breeze, s_fo and s_ob must not be None")
            fo_s = new_Fo_s[line,:]
            index = np.where((s_ob >= 5.5)|(fo_s>=5.5))
            breeze_count =d_ob.size - len(index[0])
            ob = ob0[index]
            fo = fo[index]
        else:
            ob = ob0
            breeze_count = 0
        nasd_array = np.zeros(3)
        nasd_array[0] = d_ob.size
        d_angle = np.abs(fo - ob)
        d_angle[d_angle == 6] = 1
        index = np.where(d_angle ==0)
        nasd_array[1] = len(index[0]) + breeze_count
        nasd_array[2] = len(index[0]) + breeze_count
        index = np.where(d_angle == 1)
        nasd_array[2] += len(index[0]) * 0.6

        nasd_list.append(nasd_array)
    nasd_array = np.array(nasd_list)
    shape = list(Fo_shape[:ind])
    shape.append(3)
    nasd_array = nasd_array.reshape(shape)
    return nasd_array

def scd(d_ob,d_fo,s_ob = None,s_fo = None, ignore_breeze = False):
    '''
    计算风向预报评分。
    基于原始风向数组，计算风向预报评分。计算的第一步是将预报和观测的风向都转换成8个离散的方位角，
    当一个样本的预报观测方位角正好相同时，得1分，方位角差1级，得0.6分，否则得0分。风向评分等于所有样本得分的平均。
    :param d_ob:观测的风向，numpy数组
    :param s_ob:观测的风速，numpy数组，shape和d_ob完全一致
    :param d_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param s_fo:预报的风速，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param ignore_breeze ： 该参数为True时，若预报和观测的风速都小于等于3级，则认为风向是正确的。
    :return: 风向预报评分，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    nasd_array = nas_d(d_ob,d_fo,s_ob = s_ob,s_fo = s_fo,ignore_breeze= ignore_breeze)
    scd0 = scd_nas(nasd_array)
    return scd0

def acd(d_ob,d_fo,s_ob = None,s_fo = None, ignore_breeze = False,unit = 1):
    '''
    计算风向预报准确率。
    基于原始风向数组，计算风向预报准确率，计算的第一步是将预报和观测的风向都转换成8个离散的方位角，
    当预报和观测的方位角正好相同时，就记为正确，否则记为错误，风向预报准确率时正确样本数/总样本数。
    :param d_ob:观测的风向，numpy数组
    :param s_ob:观测的风速，numpy数组，shape和d_ob完全一致
    :param d_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param s_fo:预报的风速，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param ignore_breeze ： 该参数为True时，若预报和观测的风速都小于等于3级，则认为风向是正确的。
    :return: 风向预报准确率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    nasd_array = nas_d(d_ob,d_fo,s_ob = s_ob,s_fo = s_fo,ignore_breeze= ignore_breeze)
    acd0 = acd_nas(nasd_array,unit=unit)
    return acd0

def nas_uv(u_ob,u_fo,v_ob,v_fo, ignore_breeze = False):
    '''
    根据u，v分量计算风向预报准确率和评分的中间量。
    将输入的观测和预报风向（0-360度），转换成计算风向预报准确率，风向预报评分所需要的中间量
    ，计算的第一步是将预报和观测的风向都转换成8个离散的方位角。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :param ignore_breeze:  该参数为True时，若预报和观测的风速都小于等于3级，则认为风向是正确的。
    :return: 根据预报和观测的风向数据，统计得到的样本数、风向正确样本数、风向评分（分子部分），d_fo和d_ob的shape一致，则说明只有一家预报，此时返回
             一维的数据，size= 3。 如果d_fo和d_ob的shape不一致，则说明有多个预报，则返回的数据shape = （预报成员数，3）
    '''
    s_ob,d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob,v_ob)
    s_fo,d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo,v_fo)
    return nas_d(d_ob,d_fo,s_ob = s_ob,s_fo = s_fo,ignore_breeze= ignore_breeze)

def scd_uv(u_ob,u_fo,v_ob,v_fo, ignore_breeze = False):
    '''
    计算风向预报评分。
    基于原始风向数组，计算风向预报评分。计算的第一步是将预报和观测的风向都转换成8个离散的方位角，
    当一个样本的预报观测方位角正好相同时，得1分，方位角差1级，得0.6分，否则得0分。风向评分等于所有样本得分的平均。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :param ignore_breeze ： 该参数为True时，若预报和观测的风速都小于等于3级，则认为风向是正确的。
    :return: 风向预报评分，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob,d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob,v_ob)
    s_fo,d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo,v_fo)
    #print(d_ob)
    nasd_array = nas_d(d_ob, d_fo, s_ob=s_ob, s_fo=s_fo, ignore_breeze=ignore_breeze)
    scd0 = scd_nas(nasd_array)
    return scd0

def acd_uv(u_ob,u_fo,v_ob,v_fo, ignore_breeze = False,unit = 1):
    '''
    计算风向预报准确率。
    基于u,v风分量，计算风向预报准确率。计算的第一步是将预报和观测的风向都转换成8个离散的方位角，
    当一个样本的预报观测方位角正好相同时，得1分，方位角差1级，得0.6分，否则得0分。风向评分等于所有样本得分的平均。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的v分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :param ignore_breeze ： 该参数为True时，若预报和观测的风速都小于等于3级，则认为风向是正确的。
    :return: 风向预报准确率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob,d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob,v_ob)
    s_fo,d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo,v_fo)
    nasd_array = nas_d(d_ob, d_fo, s_ob=s_ob, s_fo=s_fo, ignore_breeze=ignore_breeze)
    acd0 = acd_nas(nasd_array,unit=unit)
    return acd0


def nasws_s(s_ob,s_fo,min_s = 0,max_s = 300):
    '''
    将输入的观测和预报风速（m/s），转换成计算风速预报准确率，风向预报评分所需要的中间量
    :param s_ob:观测的风向，numpy数组，计算的第一步是将预报和观测的风速都转换成14个离散的等级，
    :param s_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return:根据预报和观测的风速数据，统计得到的样本数、风速等级正确样本数、风速评分（分子部分）,偏弱样本数，偏强样本数，d_fo和d_ob的shape一致，则说明只有一家预报，此时返回
             一维的数据，size= 5。 如果d_fo和d_ob的shape不一致，则说明有多个预报，则返回的数据shape = （预报成员数，5）
    '''
    Ob_shape = s_ob.shape
    Fo_shape = s_fo.shape
    nass_list = []
    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = s_fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape

    ob0 = meteva.base.tool.math_tools.tran_speed_to_14grade(s_ob)
    #print(ob0)
    for line in range(new_Fo_shape[0]):
        fo = new_Fo[line, :]
        #print(fo)
        if min_s>0 or max_s < 300:
            index = np.where(((s_ob>=min_s) & (s_ob < max_s)) | ((fo >= min_s) & (fo < max_s)))
            ob = ob0[index]
            fo = fo[index]
        else:
            ob = ob0

        fo = meteva.base.tool.math_tools.tran_speed_to_14grade(fo)
        nass_array = np.zeros(5)
        d_grade = np.abs(fo - ob)
        nass_array[0] = ob.size
        index = np.where(d_grade ==0)
        nass_array[1] = len(index[0])
        nass_array[2] = len(index[0])
        index = np.where(d_grade == 1)
        nass_array[2] += len(index[0]) * 0.6
        index = np.where(d_grade == 2)
        nass_array[2] += len(index[0]) * 0.4
        index = np.where(fo < ob)
        nass_array[3] += len(index[0])
        index = np.where(fo > ob)
        nass_array[4] += len(index[0])
        nass_list.append(nass_array)


    nass_array = np.array(nass_list)
    shape = list(Fo_shape[:ind])
    shape.append(5)
    nass_array = nass_array.reshape(shape)



    return nass_array


def nasws_uv(u_ob,u_fo,v_ob,v_fo,min_s = 0,max_s = 300):
    '''
    根据u，v分量计算风速预报准确率和评分的中间量。
    将输入的观测和预报风向（0-360度），转换成计算风向预报准确率，风向预报评分所需要的中间量
    ，计算的第一步是将预报和观测的风向都转换成8个离散的方位角。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 根据预报和观测的风向数据，统计得到的样本数、风向正确样本数、风向评分（分子部分），d_fo和d_ob的shape一致，则说明只有一家预报，此时返回
             一维的数据，size= 3。 如果d_fo和d_ob的shape不一致，则说明有多个预报，则返回的数据shape = （预报成员数，3）
    '''
    s_ob,d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob,v_ob)
    s_fo,d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo,v_fo)
    return nasws_s(s_ob,s_fo,min_s = min_s,max_s = max_s)


def scs_nasws(nass_array):
    '''
    基于中间结果计算风速预报评分
    :param nasd_array: 输入nass函数统计得到的样本数、风速等级正确样本数、风速预报评分（分子部分），偏弱样本数，偏强样本数
    :return: 返回平均的风速预报评分
    '''
    total = nass_array[...,0]
    sc = nass_array[...,2]
    scs = sc/total
    return scs

def scs(s_ob,s_fo,min_s = 0,max_s = 300):
    '''
    基于原始风速（m/s)数组，计算风速预报评分。计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当一个样本的预报和观测风速等级正好相同时，得1分，等级差1级，得0.6分，等级差2级得0.4分,否则不得分。风速评分等于所有样本得分的平均。
    :param s_ob:观测的风向，numpy数组，
    :param s_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 风向预报评分，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    nass_array = nasws_s(s_ob,s_fo,min_s = min_s,max_s =max_s)
    scs0 = scs_nasws(nass_array)
    return scs0

def scs_uv(u_ob, u_fo, v_ob, v_fo, min_s = 0,max_s = 300):
    '''
    风速预报评分。
    基于原始u，v风（m/s)数组，计算风速预报评分。计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当一个样本的预报和观测风速等级正好相同时，得1分，等级差1级，得0.6分，等级差2级得0.4分,否则不得分。风速评分等于所有样本得分的平均。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 风速预报评分，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob, d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob, v_ob)
    s_fo, d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo, v_fo)
    nass_array = nasws_s(s_ob,s_fo,min_s = min_s,max_s =max_s)
    scs0 = scs_nasws(nass_array)
    return scs0


def acs_nasws(nass_array,unit = 1):
    '''
    基于中间结果计算风速预报准确率，
    :param nasd_array: 输入nasd函数统计得到的样本数、风速正确样本数、风速评分（分子部分），偏弱样本数，偏强样本数
    :return: 返回平均的风速预报准确率
    '''
    total = nass_array[...,0]
    sc = nass_array[...,1]
    acs = sc/total
    if unit =="%":
        acs *= 100
    return acs

def acs(s_ob,s_fo,min_s = 0,max_s = 300,unit = 1):
    '''
    基于原始风速（m/s)数组，计算风速预报准确率。计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当一个样本的预报和观测风速等级正好相同时，记为正确，否则记为错误。风速评分等于所有样本得分的平均。
    :param s_ob:观测的风向，numpy数组，
    :param s_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 风速预报准确率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    nass_array = nasws_s(s_ob, s_fo,min_s = min_s,max_s =max_s)
    acs0 = acs_nasws(nass_array,unit = unit)
    return acs0


def acs_uv(u_ob, u_fo, v_ob, v_fo, min_s = 0,max_s = 300,unit = 1):
    '''
    风速预报准确率。
    基于原始u，v风（m/s)数组，计算风速预报准确率。计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当一个样本的预报和观测风速等级正好相同时，记为正确，否则记为错误。风速评分等于所有样本得分的平均。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 风速预报准确率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob, d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob, v_ob)
    s_fo, d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo, v_fo)
    nass_array = nasws_s(s_ob, s_fo,min_s = min_s,max_s =max_s)
    acs0 = acs_nasws(nass_array,unit = unit)
    return acs0


def wind_severer_rate_nasws(nasws_array,unit = 1):
    '''
    基于中间结果计算风速预报偏强率，
    :param nasd_array: 输入nasd函数统计得到的样本数、风速正确样本数、风速评分（分子部分），偏弱样本数，偏强样本数
    :return: 返回平均的风速预报准确率
    '''
    total = nasws_array[...,0]
    sc = nasws_array[...,4]
    rs = sc/total
    if unit == "%":
        rs *= 100
    return rs

def wind_severer_rate(s_ob,s_fo,min_s = 0,max_s = 300,unit = 1):
    '''
    基于原始风向数组，计算风速预报偏强率，计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当预报风速等级大于观测风速等级时记为1，否则记为0，偏强率是偏强的样本数/总样本数
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 风向预报准确率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    nass_array = nasws_s(s_ob, s_fo,min_s = min_s,max_s =max_s)
    rs = wind_severer_rate_nasws(nass_array,unit = unit)
    return rs

def wind_severer_rate_uv(u_ob, u_fo, v_ob, v_fo, min_s = 0,max_s = 300,unit = 1):
    '''
    基于原始u，v风（m/s)数组，计算风速预报偏强率，计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当预报风速等级大于观测风速等级时记为1，否则记为0，偏强率是偏强的样本数/总样本数
    :param s_ob:观测的风向，numpy数组，
    :param s_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 风向预报准确率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob, d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob, v_ob)
    s_fo, d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo, v_fo)
    nass_array = nasws_s(s_ob, s_fo,min_s = min_s,max_s =max_s)
    rs = wind_severer_rate_nasws(nass_array,unit = unit)
    return rs

def wind_weaker_rate_nasws(nasws_array,unit = 1):
    '''
    基于中间结果计算风速预报偏弱率，
    :param nasd_array: 输入nasd函数统计得到的样本数、风速正确样本数、风速评分（分子部分），偏弱样本数，偏强样本数
    :return: 风速预报偏弱率
    '''
    total = nasws_array[...,0]
    sc = nasws_array[...,3]
    rw = sc/total
    if unit =="%":
        rw *= 100
    return rw

def wind_weaker_rate(s_ob,s_fo,min_s = 0,max_s = 300,unit = 1):
    '''
    基于原始风向数组，计算风速预报偏弱率，计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当预报风速等级小于观测风速等级时记为1，否则记为0，偏强率是偏弱的样本数/总样本数
    :param s_ob:观测的风向，numpy数组，
    :param s_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 风速预报偏弱率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    nass_array = nasws_s(s_ob, s_fo,min_s = min_s,max_s =max_s)
    rw = wind_weaker_rate_nasws(nass_array,unit=unit)
    return rw


def wind_weaker_rate_uv(u_ob, u_fo, v_ob, v_fo, min_s = 0,max_s = 300,unit = 1):
    '''
    基于原始风向数组，计算风速预报偏弱率，计算的第一步是将预报和观测的风向都转换成14个离散的风速等级，
    当预报风速等级小于观测风速等级时记为1，否则记为0，偏强率是偏弱的样本数/总样本数
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :param min_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :param max_s:  过滤参数，当样本的预报或者观测的风速位于[min_s,max_s)区间时，会将样本纳入统计样本，否则会将样本删除
    :return: 风速预报偏弱率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob, d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob, v_ob)
    s_fo, d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo, v_fo)
    nass_array = nasws_s(s_ob, s_fo,min_s = min_s,max_s =max_s)
    rw = wind_weaker_rate_nasws(nass_array,unit = unit)
    return rw


def na_ds(d_ob,d_fo,s_ob,s_fo):
    '''
    计算风预报综合检验所需的中间量。
        将输入的观测和预报风向（0-360度）和风速（m/s），转换成计算风预报综合准确率所需要的中间量，
        ，计算的第一步是将预报和观测的风向都转换成8个离散的方位角，将观测和预报的风速转换成等级
        :param d_ob:观测的风向，numpy数组
        :param s_ob:观测的风速，numpy数组，shape和d_ob完全一致
        :param d_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
        :param s_fo:预报的风速，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
        :return:根据预报和观测的风向风速数据，统计得到的样本数、风向和风速都正确样本数，d_fo和d_ob的shape一致，则说明只有一家预报，此时返回
                 一维的数据，size= 2。 如果d_fo和d_ob的shape不一致，则说明有多个预报，则返回的数据shape = （预报成员数，2）
        '''
    Ob_shape = d_ob.shape
    Fo_shape = d_fo.shape
    nasd_list = []
    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo_d = d_fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo_d.shape
    new_Fo_s = s_fo.reshape(new_Fo_shape)


    ob_d = meteva.base.tool.math_tools.tran_direction_to_8angle(d_ob)
    ob_s = meteva.base.tool.math_tools.tran_speed_to_14grade(s_ob)
    for line in range(new_Fo_shape[0]):
        fo_d = meteva.base.tool.math_tools.tran_direction_to_8angle(new_Fo_d[line, :])
        fo_s = meteva.base.tool.math_tools.tran_speed_to_14grade(new_Fo_s[line,:])
        nasd_array = np.zeros(2)
        nasd_array[0] = d_ob.size
        index = np.where((fo_d == ob_d)&(fo_s == ob_s))
        nasd_array[1] += len(index[0])
        nasd_list.append(nasd_array)
    nrag_array = np.array(nasd_list)
    shape = list(Fo_shape[:ind])
    shape.append(2)
    nrag_array = nrag_array.reshape(shape)
    return nrag_array


def na_uv(u_ob,u_fo,v_ob,v_fo):

    '''
    将输入的观测和预报u,v分量（m/s)，转换成计算风向预报准确率，风向预报评分所需要的中间量
    ，计算的第一步是将预报和观测的风向都转换成8个离散的方位角。
    :param u_ob: 观测的u分量，numpy数组
    :param u_fo: 预报的u分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），u_fo.shape低维与u_ob.shape保持一致
    :param v_ob: 观测的u分量，numpy数组，shape 和u_ob完全一致
    :param v_fo: 预报的v分量，numpy数组，shape和u_ob完全一致或比u_ob高一维（用于同时进行多家预报结果检验），v_fo.shape低维与v_ob.shape保持一致
    :return:根据预报和观测的风向风速数据，统计得到的样本数、风向和风速都正确样本数，d_fo和d_ob的shape一致，则说明只有一家预报，此时返回
                 一维的数据，size= 2。 如果d_fo和d_ob的shape不一致，则说明有多个预报，则返回的数据shape = （预报成员数，2）
    '''
    s_ob,d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob,v_ob)
    s_fo,d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo,v_fo)
    return na_ds(d_ob,d_fo,s_ob,s_fo)


def acz_na(na_array):
    '''
    基于中间结果计算风预报综合准确率，
    :param nrag_array: 输入nrag函数统计得到的样本数、风向风速都正确的样本数
    :return: 返回风预报综合准确率
    '''
    total = na_array[...,0]
    sc = na_array[...,1]
    ac = sc/total
    return ac

def acz(d_ob,d_fo,s_ob,s_fo):
    '''
    根据输入的观测和预报风向（0-360度）和风速（m/s），计算风雨吧综合准确率，
    ，计算的第一步是将预报和观测的风向都转换成8个离散的方位角，将观测和预报的风速转换成等级，
    观测预报的风向方位角一致且风速等级一致则记为正确，否则记为错误，风预报综合准确率= 正确样本数/总样本数
    :param d_ob:观测的风向，numpy数组
    :param s_ob:观测的风速，numpy数组，shape和d_ob完全一致
    :param d_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param s_fo:预报的风速，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :return:返回风预报综合准确率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    narg_array = na_ds(d_ob,d_fo,s_ob,s_fo)
    acz1 = acz_na(narg_array)
    return acz1

def acz_uv(u_ob,u_fo,v_ob,v_fo):
    '''
    根据输入的观测和预报u,v分量（m/s)，计算风预报综合准确率，
    ，计算的第一步是将预报和观测的风向都转换成8个离散的方位角，将观测和预报的风速转换成等级，
    观测预报的风向方位角一致且风速等级一致则记为正确，否则记为错误，风预报综合准确率= 正确样本数/总样本数
    :param d_ob:观测的风向，numpy数组
    :param s_ob:观测的风速，numpy数组，shape和d_ob完全一致
    :param d_fo:预报的风向，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :param s_fo:预报的风速，numpy数组，numpy可以比d_ob高一维（用于同时进行多家预报结果检验），fo.shape低维与ob.shape保持一致
    :return:返回风预报综合准确率，如果d_fo和d_ob的shape一致，说明只有一家预报，则返回实数，否则说明是在同时检验多家预报，返回结果为一维数组。
    '''
    s_ob,d_ob = meteva.base.math_tools.u_v_to_s_d(u_ob,v_ob)
    s_fo,d_fo = meteva.base.math_tools.u_v_to_s_d(u_fo,v_fo)
    narg_array = na_ds(d_ob,d_fo,s_ob,s_fo)
    acz1 = acz_na(narg_array)
    return acz1


def tdis(lon_ob,lat_ob,lon_fo,lat_fo,on_earth_surface = True):
    '''

    :param x_ob:
    :param y_ob:
    :param x_fo:
    :param y_fo:
    :return:
    '''
    Ob_shape = lon_ob.shape
    Fo_shape = lon_fo.shape
    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    x_new_Fo = lon_fo.reshape(new_Fo_shape)
    y_new_Fo = lat_fo.reshape(new_Fo_shape)
    new_Fo_shape = x_new_Fo.shape

    # print(ob0)
    dis_list = []
    total_count = lon_ob.size
    for line in range(new_Fo_shape[0]):
        fox = x_new_Fo[line, :]
        foy = y_new_Fo[line, :]
        if on_earth_surface:
            dis_array = meteva.base.math_tools.distance_on_earth_surface(lon_ob,lat_ob,fox,foy)
        else:
            dis2 = (fox - lon_ob)*(fox - lon_ob) + (foy - lat_ob)*(foy - lat_ob)
            dis_array = np.sqrt(dis2)
        dis_sum = np.sum(dis_array)
        dis_list.append([total_count,dis_sum])

    tdis_np = np.array(dis_list)
    shape = list(Fo_shape[:ind])
    shape.append(2)
    tdis_array = tdis_np.reshape(shape)
    return tdis_array


def distance_tdis(tdis_array):
    '''
    me 求两组数据的误差平均值
    :param tase_array:包含命中空报和漏报的多维数组，其中最后一维长度为4，分别记录了（total_count,e_sum,ae_sum,se_sum）
    （样本数，误差和、绝对误差和，误差平方和），它由tase返回
    :return: 负无穷到正无穷的实数，最优值为0
    '''
    mean_dis = tdis_array[..., 1] / tdis_array[..., 0]
    return mean_dis


def distance(lon_ob,lat_ob,lon_fo,lat_fo,on_earth_surface = True):
    '''

    :param x_ob:
    :param y_ob:
    :param x_fo:
    :param y_fo:
    :param on_earth:
    :return:
    '''
    tdis_array = tdis(lon_ob,lat_ob,lon_fo,lat_fo,on_earth_surface= on_earth_surface)
    mean_dis = distance_tdis(tdis_array)
    return mean_dis



